from pathlib import Path

PRODUCT_NAME = "lakebridge"
PRODUCT_PATH_PREFIX = Path.home() / ".databricks" / "labs" / PRODUCT_NAME / "lib"

PLATFORM_TO_SOURCE_TECHNOLOGY_CFG = {
    "synapse": "src/databricks/labs/lakebridge/resources/assessments/synapse/pipeline_config.yml",
}

# TODO modify this PLATFORM_TO_SOURCE_TECHNOLOGY.keys() once all platforms are supported
PROFILER_SOURCE_SYSTEM = ["synapse"]


# This flag indicates whether a connector is required for the source system when pipeline is trigger
# For example in the case of synapse no connector is required and the python scripts
# manage the connection by directly reading the credentials files
# Revisit this when more source systems are added to standardize the approach
CONNECTOR_REQUIRED = {
    "synapse": False,
    "mssql": True,
}



import logging
import time

from collections.abc import Iterator

from databricks.labs.blueprint.tui import Prompts
from databricks.sdk import WorkspaceClient
from databricks.sdk.service.catalog import Privilege, SecurableType
from databricks.sdk.service.sql import (
    CreateWarehouseRequestWarehouseType,
    EndpointInfoWarehouseType,
    SpotInstancePolicy,
)
from databricks.sdk.service.serving import ServingEndpoint

from databricks.labs.lakebridge.helpers.metastore import CatalogOperations

logger = logging.getLogger(__name__)


class ResourceConfigurator:
    """
    Handles the setup of common Databricks resources like
    catalogs, schemas, volumes, and warehouses used across remorph modules.
    """

    def __init__(self, ws: WorkspaceClient, prompts: Prompts, catalog_ops: CatalogOperations):
        self._ws = ws
        self._user = ws.current_user.me()
        self._prompts = prompts
        self._catalog_ops = catalog_ops

    def prompt_for_catalog_setup(
        self,
        default_catalog_name: str = "remorph",
    ) -> str:
        catalog_name = self._prompts.question("Enter catalog name", default=default_catalog_name)
        catalog = self._catalog_ops.get_catalog(catalog_name)
        if catalog:
            logger.info(f"Found existing catalog `{catalog_name}`")
            return catalog_name
        if self._prompts.confirm(f"Catalog `{catalog_name}` doesn't exist. Create it?"):
            result = self._catalog_ops.create_catalog(catalog_name)
            assert result.name is not None
            return result.name
        raise SystemExit("Cannot continue installation, without a valid catalog, Aborting the installation.")

    def prompt_for_schema_setup(
        self,
        catalog: str,
        default_schema_name: str,
    ) -> str:
        schema_name = self._prompts.question("Enter schema name", default=default_schema_name)
        schema = self._catalog_ops.get_schema(catalog, schema_name)
        if schema:
            logger.info(f"Found existing schema `{schema_name}` in catalog `{catalog}`")
            return schema_name
        if self._prompts.confirm(f"Schema `{schema_name}` doesn't exist in catalog `{catalog}`. Create it?"):
            result = self._catalog_ops.create_schema(schema_name, catalog)
            assert result.name is not None
            return result.name
        raise SystemExit("Cannot continue installation, without a valid schema. Aborting the installation.")

    def prompt_for_volume_setup(
        self,
        catalog: str,
        schema: str,
        default_volume_name: str,
    ) -> str:
        volume_name = self._prompts.question("Enter volume name", default=default_volume_name)
        volume = self._catalog_ops.get_volume(catalog, schema, volume_name)
        if volume:
            logger.info(f"Found existing volume `{volume_name}` in catalog `{catalog}` and schema `{schema}`")
            return volume_name
        if self._prompts.confirm(
            f"Volume `{volume_name}` doesn't exist in catalog `{catalog}` and schema `{schema}`. Create it?"
        ):
            result = self._catalog_ops.create_volume(catalog, schema, volume_name)
            assert result.name is not None
            return result.name
        raise SystemExit("Cannot continue installation, without a valid volume. Aborting the installation.")

    def prompt_for_warehouse_setup(self, warehouse_name_prefix: str) -> str:
        def warehouse_type(_):
            return _.warehouse_type.value if not _.enable_serverless_compute else "SERVERLESS"

        pro_warehouses = {"[Create new PRO SQL warehouse]": "create_new"} | {
            f"{_.name} ({_.id}, {warehouse_type(_)}, {_.state.value})": _.id
            for _ in self._ws.warehouses.list()
            if _.state is not None and _.warehouse_type == EndpointInfoWarehouseType.PRO
        }
        warehouse_id = self._prompts.choice_from_dict(
            "Select PRO or SERVERLESS SQL warehouse",
            pro_warehouses,
        )
        if warehouse_id == "create_new":
            new_warehouse = self._ws.warehouses.create(
                name=f"{warehouse_name_prefix} {time.time_ns()}",
                spot_instance_policy=SpotInstancePolicy.COST_OPTIMIZED,
                warehouse_type=CreateWarehouseRequestWarehouseType.PRO,
                cluster_size="Small",
                max_num_clusters=1,
            )
            warehouse_id = new_warehouse.id

        if warehouse_id is None:
            raise SystemExit("Cannot continue installation, without a valid warehouse. Aborting the installation.")
        return warehouse_id

    def prompt_for_foundation_model_choice(self, default_choice: str = "databricks-claude-sonnet-4-5") -> str:
        """
        List Serving Endpoints that expose a foundation model and prompt the user to pick one.
        Returns the selected endpoint name
        """
        endpoints: Iterator[ServingEndpoint] = self._ws.serving_endpoints.list()

        model_endpoints = [
            ep
            for ep in endpoints
            if ep.name
            and ep.config
            and ep.config.served_entities
            and any(getattr(se, "foundation_model", None) is not None for se in ep.config.served_entities)
        ]

        foundation_model_names = [ep.name for ep in model_endpoints if ep.name]

        if not foundation_model_names:
            raise RuntimeError("No Foundation Model serving endpoints found. Aborting the installation.")
        # This logic is implemented to make the default choice always to appear first in the list
        other_models = sorted(set(foundation_model_names) - {default_choice})
        choices = [f"[Recommended] {default_choice}", *other_models]
        selected = self._prompts.choice("Select a Foundation Model serving endpoint:", choices, sort=False)

        if selected.startswith("[Recommended]"):
            selected = default_choice
        return selected

    def has_necessary_catalog_access(
        self, catalog_name: str, user_name: str, privilege_sets: tuple[set[Privilege], ...]
    ):
        catalog = self._catalog_ops.get_catalog(catalog_name)
        assert catalog, f"Catalog not found {catalog_name}"
        if self._catalog_ops.has_catalog_access(catalog, user_name, privilege_sets):
            return True
        missing_permissions = self._get_missing_permissions(
            user_name, SecurableType.CATALOG, catalog.name, privilege_sets
        )
        logger.error(
            f"User `{user_name}` doesn't have required privileges :: \n`{missing_permissions}`\n to access catalog `{catalog_name}` "
        )
        return False

    def has_necessary_schema_access(
        self, catalog_name: str, schema_name: str, user_name: str, privilege_sets: tuple[set[Privilege], ...]
    ):
        schema = self._catalog_ops.get_schema(catalog_name, schema_name)
        assert schema, f"Schema not found {catalog_name}.{schema_name}"
        if self._catalog_ops.has_schema_access(schema, user_name, privilege_sets):
            return True
        missing_permissions = self._get_missing_permissions(
            user_name, SecurableType.SCHEMA, schema.full_name, privilege_sets
        )
        logger.error(
            f"User `{user_name}` doesn't have required privileges :: \n`{missing_permissions}`\n to access schema `{schema.full_name}` "
        )
        return False

    def has_necessary_volume_access(
        self,
        catalog_name: str,
        schema_name: str,
        volume_name: str,
        user_name: str,
        privilege_sets: tuple[set[Privilege], ...],
    ):
        volume = self._catalog_ops.get_volume(catalog_name, schema_name, volume_name)
        assert volume, f"Volume not found {catalog_name}.{schema_name}.{volume_name}"
        if self._catalog_ops.has_volume_access(volume, user_name, privilege_sets):
            return True
        missing_permissions = self._get_missing_permissions(
            user_name, SecurableType.VOLUME, volume.full_name, privilege_sets
        )
        logger.error(
            f"User `{user_name}` doesn't have required privileges :: \n`{missing_permissions}`\n to access volume `{volume.full_name}` "
        )
        return False

    def _get_missing_permissions(
        self,
        user_name: str,
        securable_type: SecurableType,
        resource_name: str | None,
        privilege_sets: tuple[set[Privilege], ...],
    ):
        assert resource_name, f"Catalog Resource name must be provided {resource_name}"
        missing_permissions_list = []
        for privilege_set in privilege_sets:
            permissions = self._catalog_ops.has_privileges(user_name, securable_type, resource_name, privilege_set)
            if not permissions:
                missing_privileges = ", ".join([privilege.name for privilege in privilege_set])
                missing_permissions_list.append(f" * '{missing_privileges}' ")

        return " OR \n".join(missing_permissions_list)

    def has_necessary_access(self, catalog_name: str, schema_name: str, volume_name: str | None):
        catalog_required_privileges: tuple[set[Privilege], ...] = (
            {Privilege.ALL_PRIVILEGES},
            {Privilege.USE_CATALOG},
        )
        schema_required_privileges: tuple[set[Privilege], ...] = (
            {Privilege.ALL_PRIVILEGES},
            {Privilege.USE_SCHEMA, Privilege.MODIFY, Privilege.SELECT, Privilege.CREATE_VOLUME},
            {Privilege.USE_SCHEMA, Privilege.MODIFY, Privilege.SELECT},
        )
        volume_required_privileges: tuple[set[Privilege], ...] = (
            {Privilege.ALL_PRIVILEGES},
            {Privilege.READ_VOLUME, Privilege.WRITE_VOLUME},
        )

        user_name = self._user.user_name
        assert user_name is not None

        catalog_access = self.has_necessary_catalog_access(catalog_name, user_name, catalog_required_privileges)
        schema_access = self.has_necessary_schema_access(
            catalog_name, schema_name, user_name, schema_required_privileges
        )
        required_access = catalog_access and schema_access
        if volume_name:
            volume_access = self.has_necessary_volume_access(
                catalog_name, schema_name, volume_name, user_name, volume_required_privileges
            )
            required_access = required_access and volume_access
        if not required_access:
            raise SystemExit("Cannot continue installation, without necessary access. Aborting the installation.")



import json
import sys
import zoneinfo
from datetime import date, timedelta

import pandas as pd
from databricks.labs.blueprint.entrypoint import get_logger

from databricks.labs.lakebridge import initialize_logging
from databricks.labs.lakebridge.assessments import PRODUCT_NAME
from databricks.labs.lakebridge.connections.credential_manager import create_credential_manager
from databricks.labs.lakebridge.connections.env_getter import EnvGetter
from databricks.labs.lakebridge.resources.assessments.synapse.common.duckdb_helpers import insert_df_to_duckdb
from databricks.labs.lakebridge.resources.assessments.synapse.common.functions import (
    arguments_loader,
    create_synapse_artifacts_client,
)
from databricks.labs.lakebridge.resources.assessments.synapse.common.profiler_classes import SynapseWorkspace


logger = get_logger(__file__)


def execute():
    db_path, creds_file = arguments_loader(desc="Workspace Extract")

    cred_manager = create_credential_manager(PRODUCT_NAME, EnvGetter())
    synapse_workspace_settings = cred_manager.get_credentials("synapse")
    tz_info = synapse_workspace_settings["workspace"]["tz_info"]
    workspace_tz = zoneinfo.ZoneInfo(tz_info)
    workspace_name = synapse_workspace_settings["workspace"]["name"]

    logger.info(f"workspace_name: {workspace_name}")

    artifacts_client = create_synapse_artifacts_client(synapse_workspace_settings)

    try:
        # Initialize workspace settings and client
        workspace = SynapseWorkspace(workspace_tz, artifacts_client)

        # Extract workspace info
        table_name = "workspace_workspace_info"
        logger.info(f"Extraction started for {table_name}")
        workspace_info = workspace.get_workspace_info()
        workspace_info_df = pd.json_normalize([workspace_info])
        insert_df_to_duckdb(workspace_info_df, db_path, table_name)

        # Extract SQL pools
        table_name = "workspace_sql_pools"
        logger.info(f"Extraction started for {table_name}")
        sql_pools = workspace.list_sql_pools()
        sql_pools_df = pd.json_normalize([pool for pool_pages in sql_pools for pool in pool_pages])
        insert_df_to_duckdb(sql_pools_df, db_path, table_name)

        # Extract Spark pools
        table_name = "workspace_spark_pools"
        logger.info(f"Extraction started for {table_name}")
        spark_pools = workspace.list_bigdata_pools()
        spark_pools_df = pd.json_normalize([pool for pool_pages in spark_pools for pool in pool_pages])
        insert_df_to_duckdb(spark_pools_df, db_path, table_name)

        # Extract Linked Services
        table_name = "workspace_linked_services"
        logger.info(f"Extraction started for {table_name}")
        linked_services = workspace.list_linked_services()
        linked_services_list = [svc for svc_pages in linked_services for svc in svc_pages]
        linked_services_df = pd.json_normalize(linked_services_list) if linked_services_list else pd.DataFrame()
        insert_df_to_duckdb(linked_services_df, db_path, table_name)

        # Extract Data Flows
        table_name = "workspace_dataflows"
        logger.info(f"Extraction started for {table_name}")
        dataflows = workspace.list_data_flows()
        dataflows_list = [flow for flow_pages in dataflows for flow in flow_pages]
        dataflows_df = pd.json_normalize(dataflows_list) if dataflows_list else pd.DataFrame()
        insert_df_to_duckdb(dataflows_df, db_path, table_name)

        # Extract Pipelines
        table_name = "workspace_pipelines"
        logger.info(f"Extraction started for {table_name}")
        pipelines = workspace.list_pipelines()
        pipelines_list = [pipeline for pipeline_pages in pipelines for pipeline in pipeline_pages]
        pipelines_df = pd.json_normalize(pipelines_list) if pipelines_list else pd.DataFrame()
        insert_df_to_duckdb(pipelines_df, db_path, table_name)

        # Extract Spark Jobs
        table_name = "workspace_spark_jobs"
        logger.info(f"Extraction started for {table_name}")
        spark_jobs = workspace.list_spark_job_definitions()
        spark_jobs_list = [job for job_pages in spark_jobs for job in job_pages]
        spark_jobs_df = pd.json_normalize(spark_jobs_list) if spark_jobs_list else pd.DataFrame()
        insert_df_to_duckdb(spark_jobs_df, db_path, table_name)

        # Extract Notebooks
        table_name = "workspace_notebooks"
        logger.info(f"Extraction started for {table_name}")
        notebooks = workspace.list_notebooks()
        notebooks_list = [notebook for notebook_pages in notebooks for notebook in notebook_pages]
        notebooks_df = pd.json_normalize(notebooks_list) if notebooks_list else pd.DataFrame()
        insert_df_to_duckdb(notebooks_df, db_path, table_name)

        # Extract SQL Scripts
        table_name = "workspace_sql_scripts"
        logger.info(f"Extraction started for {table_name}")
        sql_scripts = workspace.list_sqlscripts()
        sql_scripts_list = [script for script_pages in sql_scripts for script in script_pages]
        sql_scripts_df = pd.json_normalize(sql_scripts_list) if sql_scripts_list else pd.DataFrame()
        insert_df_to_duckdb(sql_scripts_df, db_path, table_name)

        # Extract Triggers
        table_name = "workspace_triggers"
        logger.info(f"Extraction started for {table_name}")
        triggers = workspace.list_triggers()
        triggers_list = [trigger for trigger_pages in triggers for trigger in trigger_pages]
        triggers_df = pd.json_normalize(triggers_list) if triggers_list else pd.DataFrame()
        insert_df_to_duckdb(triggers_df, db_path, table_name)

        # Extract Libraries
        table_name = "workspace_libraries"
        logger.info(f"Extraction started for {table_name}")
        libraries = workspace.list_libraries()
        libraries_list = [lib for lib_pages in libraries for lib in lib_pages]
        libraries_df = pd.json_normalize(libraries_list) if libraries_list else pd.DataFrame()
        insert_df_to_duckdb(libraries_df, db_path, table_name)

        # Extract Datasets
        table_name = "workspace_datasets"
        logger.info(f"Extraction started for {table_name}")
        datasets = workspace.list_datasets()
        datasets_list = [dataset for dataset_pages in datasets for dataset in dataset_pages]
        datasets_df = pd.json_normalize(datasets_list) if datasets_list else pd.DataFrame()
        insert_df_to_duckdb(datasets_df, db_path, table_name)

        # Extract Pipeline Runs (last 60 days)
        today = date.today()
        pipeline_runs_list = []
        for days in range(1, 60):
            table_name = "workspace_pipeline_runs"
            logger.info(f"Extraction started for {table_name} for date: {days}")
            last_upd = today + timedelta(days=-days)
            pipeline_runs_batches = workspace.list_pipeline_runs(last_upd)
            pipeline_runs = [run for batch in pipeline_runs_batches for run in batch]
            if not pipeline_runs:
                logger.warning(f"No pipeline runs found for {last_upd}")
                continue
            if not all(isinstance(run, dict) for run in pipeline_runs):
                logger.error(f"Unexpected data export in {table_name}: {pipeline_runs}")
                raise ValueError(f"Invalid data export in {table_name}")
            for run in pipeline_runs:
                run['last_upd'] = last_upd
                pipeline_runs_list.append(run)

        pipeline_runs_df = pd.json_normalize(pipeline_runs_list)

        insert_df_to_duckdb(pipeline_runs_df, db_path, table_name)

        # Extract Trigger Runs (last 60 days)
        trigger_runs_list = []
        for days in range(1, 60):
            table_name = "workspace_trigger_runs"
            logger.info(f"Extraction started for {table_name} for date: {days}")
            last_upd = today + timedelta(days=-days)
            trigger_runs_batches = workspace.list_trigger_runs(last_upd)
            trigger_runs = [run for batch in trigger_runs_batches for run in batch]
            if not trigger_runs:
                logger.warning(f"No trigger runs found for {last_upd}")
                continue
            if not all(isinstance(run, dict) for run in trigger_runs):
                logger.error(f"Unexpected data export in {table_name}: {trigger_runs}")
                raise ValueError(f"Invalid data export in {table_name}")
            for run in trigger_runs:
                run['last_upd'] = last_upd
                trigger_runs_list.append(run)
        trigger_runs_df = pd.json_normalize(trigger_runs_list)
        insert_df_to_duckdb(trigger_runs_df, db_path, table_name)

        # This is the output format expected by the pipeline.py which orchestrates the execution of this script
        print(json.dumps({"status": "success", "message": "Data loaded successfully"}))

    except Exception as e:
        logger.error(f"Failed to extract info for Synapse Workspace Info: {str(e)}")
        print(json.dumps({"status": "error", "message": str(e)}), file=sys.stderr)
        sys.exit(1)


if __name__ == '__main__':
    initialize_logging()
    execute()



import importlib.resources
import logging
from collections.abc import Generator, Sequence
from importlib.abc import Traversable
from pathlib import PurePosixPath
from typing import Any

from databricks.labs import switch
from databricks.labs.switch.__about__ import __version__ as switch_version
from databricks.labs.blueprint.installation import Installation
from databricks.labs.blueprint.installer import InstallState
from databricks.labs.blueprint.paths import WorkspacePath
from databricks.sdk import WorkspaceClient
from databricks.sdk.errors import InvalidParameterValue, NotFound
from databricks.sdk.service import compute
from databricks.sdk.service.jobs import JobCluster, JobParameterDefinition, JobSettings, NotebookTask, Source, Task


logger = logging.getLogger(__name__)


class SwitchDeployment:
    _INSTALL_STATE_KEY = "Switch"
    _TRANSPILER_ID = "switch"

    def __init__(
        self,
        ws: WorkspaceClient,
        installation: Installation,
        install_state: InstallState,
    ):
        self._ws = ws
        self._installation = installation
        self._install_state = install_state

    def install(self, use_serverless: bool = True) -> None:
        """Deploy Switch to workspace and configure resources."""
        logger.debug("Deploying Switch resources to workspace...")
        try:
            self._deploy_resources_to_workspace()
            self._setup_job(use_serverless)
            logger.debug("Switch deployment completed")
        except (RuntimeError, ValueError, InvalidParameterValue) as e:
            msg = f"Failed to setup required resources for Switch llm transpiler: {e}"
            logger.error(msg)
            raise SystemExit(msg) from e

    def uninstall(self) -> None:
        """Remove Switch job from workspace."""
        if self._INSTALL_STATE_KEY not in self._install_state.jobs:
            logger.debug("No Switch job found in InstallState")
            return

        job_id = int(self._install_state.jobs[self._INSTALL_STATE_KEY])
        try:
            logger.info(f"Removing Switch job: id={job_id}")
            del self._install_state.jobs[self._INSTALL_STATE_KEY]
            self._ws.jobs.delete(job_id)
            self._install_state.save()
        except (InvalidParameterValue, NotFound):
            logger.debug(f"Switch job (id={job_id}) doesn't exist anymore, nothing to do.")
            self._install_state.save()

    def _get_switch_workspace_path(self) -> WorkspacePath:
        installation_root = self._installation.install_folder()
        return WorkspacePath(self._ws, installation_root) / "switch"

    def _deploy_resources_to_workspace(self) -> None:
        """Replicate the Switch package sources to the workspace."""
        # TODO: This is temporary, instead the jobs should directly run the code from the deployed wheel/package.
        resource_root = self._get_switch_workspace_path()
        # Replace existing resources, to avoid stale files and potential confusion.
        if resource_root.exists():
            resource_root.rmdir(recursive=True)
        resource_root.mkdir(parents=True)
        already_created = {resource_root}
        logger.info(f"Copying resources to workspace: {resource_root}")
        for resource_path, resource in self._enumerate_package_files(switch):
            # Resource path has a leading 'switch' that we want to strip off.
            nested_path = resource_path.relative_to(PurePosixPath("switch"))
            upload_path = resource_root / nested_path
            if (parent := upload_path.parent) not in already_created:
                logger.debug(f"Creating workspace directory: {parent}")
                parent.mkdir()
                already_created.add(parent)
            logger.debug(f"Uploading: {resource_path} -> {upload_path}")
            upload_path.write_bytes(resource.read_bytes())
        logger.debug(f"Resources copied to workspace: {resource_root}")

    @staticmethod
    def _enumerate_package_files(package) -> Generator[tuple[PurePosixPath, Traversable]]:
        # Locate the root of the package, and then enumerate all its files recursively.
        root = importlib.resources.files(package)

        def _enumerate_resources(
            resource: Traversable, parent: PurePosixPath = PurePosixPath(".")
        ) -> Generator[tuple[PurePosixPath, Traversable]]:
            if resource.name.startswith("."):
                # Skip hidden files and directories
                return
            if resource.is_dir():
                next_parent = parent / resource.name
                for child in resource.iterdir():
                    yield from _enumerate_resources(child, next_parent)
            elif resource.is_file():
                # Skip compiled Python files
                if not (name := resource.name).endswith((".pyc", ".pyo")):
                    yield parent / name, resource

        yield from _enumerate_resources(root)

    def _setup_job(self, use_serverless: bool = True) -> None:
        """Create or update Switch job."""
        existing_job_id = self._get_existing_job_id()
        logger.debug("Setting up Switch job in workspace...")
        job_id = self._create_or_update_switch_job(existing_job_id, use_serverless)
        self._install_state.jobs[self._INSTALL_STATE_KEY] = job_id
        self._install_state.save()
        job_url = f"{self._ws.config.host}/jobs/{job_id}"
        logger.info(f"Switch job created/updated: {job_url}")

    def _get_existing_job_id(self) -> str | None:
        """Check if Switch job already exists in workspace."""
        if self._INSTALL_STATE_KEY not in self._install_state.jobs:
            return None
        try:
            job_id = self._install_state.jobs[self._INSTALL_STATE_KEY]
            self._ws.jobs.get(int(job_id))
            return job_id
        except (InvalidParameterValue, NotFound, ValueError):
            return None

    def _create_or_update_switch_job(self, job_id: str | None, use_serverless: bool = True) -> str:
        """Create or update Switch job, returning job ID."""
        job_settings = self._get_switch_job_settings(use_serverless)

        # Try to update existing job
        if job_id:
            try:
                logger.debug(f"Updating Switch job (id={job_id}) with settings: {job_settings}")
                self._ws.jobs.reset(int(job_id), JobSettings(**job_settings))
                return job_id
            except (ValueError, InvalidParameterValue):
                logger.warning("Previous Switch job not found, creating new one")

        # Create new job
        logger.debug(f"Creating new Switch job with settings: {job_settings}")
        new_job = self._ws.jobs.create(**job_settings)
        new_job_id = str(new_job.job_id)
        assert new_job_id is not None
        return new_job_id

    def _get_switch_job_settings(self, use_serverless: bool = True) -> dict[str, Any]:
        """Build job settings for Switch transpiler."""
        job_name = "Lakebridge_Switch"
        notebook_path = self._get_switch_workspace_path() / "notebooks" / "00_main"

        task = Task(
            task_key="run_transpilation",
            notebook_task=NotebookTask(notebook_path=str(notebook_path), source=Source.WORKSPACE),
            disable_auto_optimization=True,  # To disable retries on failure
        )

        settings: dict[str, Any] = {
            "name": job_name,
            "tags": {"created_by": self._ws.current_user.me().user_name, "switch_version": f"v{switch_version}"},
            "tasks": [task],
            "parameters": self._generate_switch_job_parameters(),
            "max_concurrent_runs": 100,  # Allow simultaneous transpilations
        }

        if not use_serverless:
            job_cluster_key = "Switch_Cluster"
            settings["job_clusters"] = [
                JobCluster(
                    job_cluster_key=job_cluster_key,
                    new_cluster=compute.ClusterSpec(
                        spark_version=self._ws.clusters.select_spark_version(latest=True, long_term_support=True),
                        node_type_id=self._ws.clusters.select_node_type(local_disk=True, min_memory_gb=16),
                        num_workers=1,
                        data_security_mode=compute.DataSecurityMode.USER_ISOLATION,
                    ),
                )
            ]
            task.job_cluster_key = job_cluster_key

        return settings

    @staticmethod
    def _generate_switch_job_parameters() -> Sequence[JobParameterDefinition]:
        # Add required runtime parameters, static for now.
        parameters = {
            "source_tech": "",
            "input_dir": "",
            "output_dir": "",
            "foundation_model": "databricks-claude-sonnet-4-5",
            "catalog": "lakebridge",
            "schema": "switch",
        }
        return [JobParameterDefinition(name=key, default=value) for key, value in parameters.items()]



class SynapseQueries:
    @staticmethod
    def list_databases() -> str:
        """Get list of databases"""
        return """
               SELECT
                   NAME,
                   DATABASE_ID,
                   CREATE_DATE,
                   STATE,
                   STATE_DESC,
                   COLLATION_NAME
               FROM SYS.DATABASES WHERE NAME <> 'master' ;
               """

    @staticmethod
    def list_tables(pool_name: str) -> str:
        """Get list of tables"""
        return f"""
               SELECT
                   TABLE_CATALOG,
                   TABLE_SCHEMA,
                   TABLE_NAME,
                   TABLE_TYPE,
                   '{pool_name}' as POOL_NAME
               FROM INFORMATION_SCHEMA.TABLES ;
               """

    @staticmethod
    def list_columns(pool_name: str) -> str:
        """Get list of columns"""
        return f"""
               SELECT
                   TABLE_CATALOG,
                   TABLE_SCHEMA,
                   TABLE_NAME,
                   COLUMN_NAME,
                   ORDINAL_POSITION,
                   COLUMN_DEFAULT,
                   IS_NULLABLE,
                   DATA_TYPE,
                   CHARACTER_MAXIMUM_LENGTH,
                   CHARACTER_OCTET_LENGTH,
                   NUMERIC_PRECISION,
                   NUMERIC_PRECISION_RADIX,
                   NUMERIC_SCALE,
                   DATETIME_PRECISION,
                   CHARACTER_SET_CATALOG,
                   CHARACTER_SET_SCHEMA,
                   CHARACTER_SET_NAME,
                   COLLATION_CATALOG,
                   COLLATION_SCHEMA,
                   COLLATION_NAME,
                   DOMAIN_CATALOG,
                   DOMAIN_SCHEMA,
                   DOMAIN_NAME,
                   '{pool_name}' as POOL_NAME
               FROM INFORMATION_SCHEMA.COLUMNS ;
               """

    @staticmethod
    def list_views(pool_name: str) -> str:
        """Get list of views"""
        return f"""
               SELECT
                   TABLE_CATALOG,
                   TABLE_SCHEMA,
                   TABLE_NAME,
                   CHECK_OPTION,
                   IS_UPDATABLE,
                   '[REDACTED]' as VIEW_DEFINITION,
                   '{pool_name}' as POOL_NAME
               FROM INFORMATION_SCHEMA.VIEWS
               """

    @staticmethod
    def list_routines(pool_name: str) -> str:
        """Get list of routines (functions + procedures)"""
        return f"""
               SELECT
                   CREATED,
                   DATA_TYPE,
                   IS_DETERMINISTIC,
                   IS_IMPLICITLY_INVOCABLE,
                   IS_NULL_CALL,
                   IS_USER_DEFINED_CAST,
                   LAST_ALTERED,
                   MAX_DYNAMIC_RESULT_SETS,
                   NUMERIC_PRECISION,
                   NUMERIC_PRECISION_RADIX,
                   NUMERIC_SCALE,
                   ROUTINE_BODY,
                   ROUTINE_CATALOG,
                   '[REDACTED]' as ROUTINE_DEFINITION,
                   ROUTINE_NAME,
                   ROUTINE_SCHEMA,
                   ROUTINE_TYPE,
                   SCHEMA_LEVEL_ROUTINE,
                   SPECIFIC_CATALOG,
                   SPECIFIC_NAME,
                   SPECIFIC_SCHEMA,
                   SQL_DATA_ACCESS,
                   '{pool_name}' as POOL_NAME
                   FROM information_schema.routines
                   """

    @staticmethod
    def list_serverless_routines(pool_name, redact_sql_text: bool = False) -> str:
        """Get list of routines for serverless pools using sys views (information_schema.routines not supported)"""
        return f"""
               SELECT
                   SCHEMA_NAME(o.schema_id) as ROUTINE_SCHEMA,
                   o.name as ROUTINE_NAME,
                   CASE o.type
                       WHEN 'P' THEN 'PROCEDURE'
                       WHEN 'FN' THEN 'FUNCTION'
                       WHEN 'IF' THEN 'FUNCTION'
                       WHEN 'TF' THEN 'FUNCTION'
                       WHEN 'PC' THEN 'PROCEDURE'
                       ELSE o.type_desc
                   END as ROUTINE_TYPE,
                   o.create_date as CREATED,
                   o.modify_date as LAST_ALTERED,
                   '[REDACTED]' as ROUTINE_DEFINITION,
                   '{pool_name}' as POOL_NAME
               FROM sys.objects o
               WHERE o.type IN ('P', 'FN', 'IF', 'TF', 'PC')
                   AND o.is_ms_shipped = 0
               """

    @staticmethod
    def list_dedicated_sessions(pool_name: str, last_login_time: str | None = None) -> str:
        """Get session list with transformed login names and client IDs"""
        cond = "AND login_time > '" + last_login_time + "'" if last_login_time else ""
        return f"""
                     SELECT
                      APP_NAME,
                      CONVERT(VARCHAR(64), HASHBYTES('SHA2_256', CLIENT_ID), 2) as CLIENT_ID,
                      IS_TRANSACTIONAL,
                      CONVERT(VARCHAR(64), HASHBYTES('SHA2_256', LOGIN_NAME), 2) as LOGIN_NAME,
                      LOGIN_TIME,
                      QUERY_COUNT,
                      REQUEST_ID,
                      SESSION_ID,
                      SQL_SPID,
                      STATUS,
                      '{pool_name}' as POOL_NAME,
                      CURRENT_TIMESTAMP as EXTRACT_TS
                     FROM SYS.DM_PDW_EXEC_SESSIONS
                     where CHARINDEX('system', LOWER(login_name)) = 0
                         {cond}
                     """

    @staticmethod
    def list_serverless_sessions(pool_name: str, min_last_request_start_time: str | None = None) -> str:
        """Get session list with transformed login names and client IDs"""
        cond = (
            "AND last_request_start_time > '" + min_last_request_start_time + "'" if min_last_request_start_time else ""
        )
        return f"""SELECT
          ANSI_DEFAULTS,
          ANSI_NULL_DFLT_ON,
          ANSI_NULLS,
          ANSI_PADDING,
          ANSI_WARNINGS,
          ARITHABORT,
          AUTHENTICATING_DATABASE_ID,
          CLIENT_INTERFACE_NAME,
          CLIENT_VERSION,
          CONCAT_NULL_YIELDS_NULL,
          CONTEXT_INFO,
          CPU_TIME,
          DATABASE_ID,
          DATE_FIRST,
          DATE_FORMAT,
          DEADLOCK_PRIORITY,
          ENDPOINT_ID,
          GROUP_ID,
          HOST_NAME,
          HOST_PROCESS_ID,
          IS_FILTERED,
          IS_USER_PROCESS,
          LANGUAGE,
          LAST_REQUEST_END_TIME,
          LAST_REQUEST_START_TIME,
          LOCK_TIMEOUT,
          LOGICAL_READS,
          CONVERT(VARCHAR(64), HASHBYTES('SHA2_256', LOGIN_NAME), 2) as LOGIN_NAME,
          LOGIN_TIME,
          MEMORY_USAGE,
          NT_DOMAIN,
          NT_USER_NAME,
          OPEN_TRANSACTION_COUNT,
          CONVERT(VARCHAR(64), HASHBYTES('SHA2_256', ORIGINAL_LOGIN_NAME), 2) as ORIGINAL_LOGIN_NAME,
          ORIGINAL_SECURITY_ID,
          PAGE_SERVER_READS,
          PREV_ERROR,
          PROGRAM_NAME,
          QUOTED_IDENTIFIER,
          READS,
          ROW_COUNT,
          SECURITY_ID,
          SESSION_ID,
          STATUS,
          TEXT_SIZE,
          TOTAL_ELAPSED_TIME,
          TOTAL_SCHEDULED_TIME,
          TRANSACTION_ISOLATION_LEVEL,
          WRITES,
          '{pool_name}' as POOL_NAME,
          CURRENT_TIMESTAMP as EXTRACT_TS
           FROM sys.dm_exec_sessions
           WHERE is_user_process = 'True' {cond}
        """

    @staticmethod
    def list_dedicated_requests(pool_name: str, min_end_time: str | None = None, redact_sql_text: bool = True) -> str:
        """Get session request list with command type classification"""
        command_col = "'[REDACTED]' as COMMAND" if redact_sql_text else "COMMAND"
        end_time_filter = f"AND END_TIME > '{min_end_time}'" if min_end_time else ""
        return f"""
            SELECT
                CLIENT_CORRELATION_ID,
                {command_col},
                COMMAND2,
                COMMAND_TYPE,
                DATABASE_ID,
                END_COMPILE_TIME,
                END_TIME,
                ERROR_ID,
                GROUP_NAME,
                IMPORTANCE,
                [LABEL],
                REQUEST_ID,
                RESOURCE_ALLOCATION_PERCENTAGE,
                RESOURCE_CLASS,
                RESULT_CACHE_HIT,
                SESSION_ID,
                START_TIME,
                STATUS,
                SUBMIT_TIME,
                TOTAL_ELAPSED_TIME,
                '{pool_name}' AS POOL_NAME,
                CURRENT_TIMESTAMP AS EXTRACT_TS
            FROM (
                SELECT
                    *,
                    -- Extract the first word from command
                    UPPER(
                        LEFT(
                            LTRIM(command),
                            PATINDEX('%[^A-Za-z]%', LTRIM(command) + ' ') - 1
                        )
                    ) AS command_w1,

                    -- Extract the second word from command (approximation)
                    UPPER(
                        RTRIM(
                            LEFT(
                                LTRIM(
                                    RIGHT(
                                        command,
                                        LEN(command) - PATINDEX('%[^A-Za-z]%', LTRIM(command) + ' ')
                                    )
                                ),
                                PATINDEX(
                                    '%[^A-Za-z]%',
                                    LTRIM(
                                        RIGHT(command, LEN(command) - PATINDEX('%[^A-Za-z]%', LTRIM(command) + ' '))
                                    ) + ' '
                                ) - 1
                            )
                        )
                    ) AS command_w2,

                    -- Classify command type
                    CASE
                        WHEN UPPER(LEFT(LTRIM(command), PATINDEX('%[^A-Za-z]%', LTRIM(command) + ' ') - 1)) IN ('SELECT', 'WITH')
                            THEN 'QUERY'
                        WHEN UPPER(LEFT(LTRIM(command), PATINDEX('%[^A-Za-z]%', LTRIM(command) + ' ') - 1))
                            IN ('INSERT', 'UPDATE', 'MERGE', 'DELETE', 'TRUNCATE', 'COPY', 'IF', 'BEGIN', 'DECLARE', 'BUILDREPLICATEDTABLECACHE')
                            THEN 'DML'
                        WHEN UPPER(LEFT(LTRIM(command), PATINDEX('%[^A-Za-z]%', LTRIM(command) + ' ') - 1)) IN ('CREATE', 'DROP', 'ALTER')
                            THEN 'DDL'
                        WHEN UPPER(LEFT(LTRIM(command), PATINDEX('%[^A-Za-z]%', LTRIM(command) + ' ') - 1)) IN ('EXEC', 'EXECUTE')
                            THEN 'ROUTINE'
                        WHEN
                            UPPER(LEFT(LTRIM(command), PATINDEX('%[^A-Za-z]%', LTRIM(command) + ' ') - 1)) = 'BEGIN'
                            AND UPPER(RTRIM(
                                LEFT(
                                    LTRIM(
                                        RIGHT(command, LEN(command) - PATINDEX('%[^A-Za-z]%', LTRIM(command) + ' '))
                                    ),
                                    PATINDEX('%[^A-Za-z]%', LTRIM(
                                        RIGHT(command, LEN(command) - PATINDEX('%[^A-Za-z]%', LTRIM(command) + ' '))
                                    ) + ' ') - 1
                                )
                            )) IN ('TRAN', 'TRANSACTION')
                            THEN 'TRANSACTION_CONTROL'
                        WHEN
                            UPPER(LEFT(LTRIM(command), PATINDEX('%[^A-Za-z]%', LTRIM(command) + ' ') - 1)) = 'END'
                            AND UPPER(RTRIM(
                                LEFT(
                                    LTRIM(
                                        RIGHT(command, LEN(command) - PATINDEX('%[^A-Za-z]%', LTRIM(command) + ' '))
                                    ),
                                    PATINDEX('%[^A-Za-z]%', LTRIM(
                                        RIGHT(command, LEN(command) - PATINDEX('%[^A-Za-z]%', LTRIM(command) + ' '))
                                    ) + ' ') - 1
                                )
                            )) IN ('TRAN', 'TRANSACTION')
                            THEN 'TRANSACTION_CONTROL'
                        WHEN UPPER(LEFT(LTRIM(command), PATINDEX('%[^A-Za-z]%', LTRIM(command) + ' ') - 1)) IN ('COMMIT', 'ROLLBACK')
                            THEN 'TRANSACTION_CONTROL'
                        ELSE 'OTHER'
                    END AS command_type

                FROM SYS.DM_PDW_EXEC_REQUESTS
                WHERE START_TIME IS NOT NULL
                AND COMMAND IS NOT NULL
                {end_time_filter}
            ) requests
        """

    @staticmethod
    def get_db_storage_info(pool_name: str) -> str:
        """Get database storage information"""
        return f"""SELECT
                       PDW_NODE_ID AS NODE_ID,
                       (SUM(RESERVED_PAGE_COUNT) * 8) / 1024 AS RESERVEDSPACEMB,
                       (SUM(USED_PAGE_COUNT)  * 8) / 1024 AS USEDSPACEMB,
                       '{pool_name}' as POOL_NAME,
                        CURRENT_TIMESTAMP AS EXTRACT_TS
                    FROM SYS.DM_PDW_NODES_DB_PARTITION_STATS
                    GROUP BY PDW_NODE_ID
               """

    @staticmethod
    def list_serverless_requests(pool_name: str, min_start_time: str | None):
        """
        Get list of requests with start time filter
        """

        return f"""
          SELECT
              ANSI_DEFAULTS,
              ANSI_NULL_DFLT_ON,
              ANSI_NULLS,
              ANSI_PADDING,
              ANSI_WARNINGS,
              ARITHABORT,
              BLOCKING_SESSION_ID,
              COMMAND,
              CONCAT_NULL_YIELDS_NULL,
              CONNECTION_ID,
              CONTEXT_INFO,
              CPU_TIME,
              DATABASE_ID,
              DATE_FIRST,
              DATE_FORMAT,
              DEADLOCK_PRIORITY,
              DIST_STATEMENT_ID,
              DOP,
              ESTIMATED_COMPLETION_TIME,
              EXECUTING_MANAGED_CODE,
              GRANTED_QUERY_MEMORY,
              GROUP_ID,
              IS_RESUMABLE,
              LANGUAGE,
              LAST_WAIT_TYPE,
              LOCK_TIMEOUT,
              LOGICAL_READS,
              NEST_LEVEL,
              OPEN_RESULTSET_COUNT,
              OPEN_TRANSACTION_COUNT,
              PAGE_SERVER_READS,
              PERCENT_COMPLETE,
              PLAN_HANDLE,
              PREV_ERROR,
              QUERY_HASH,
              QUERY_PLAN_HASH,
              QUOTED_IDENTIFIER,
              READS,
              REQUEST_ID,
              ROW_COUNT,
              SCHEDULER_ID,
              SESSION_ID,
              SQL_HANDLE,
              START_TIME,
              STATEMENT_END_OFFSET,
              STATEMENT_START_OFFSET,
              STATUS,
              TASK_ADDRESS,
              TEXT_SIZE,
              TOTAL_ELAPSED_TIME,
              TRANSACTION_ID,
              TRANSACTION_ISOLATION_LEVEL,
              USER_ID,
              WAIT_RESOURCE,
              WAIT_TIME,
              WAIT_TYPE,
              WRITES,
              '{pool_name}' as POOL_NAME,
              CURRENT_TIMESTAMP as EXTRACT_TS
            FROM sys.dm_exec_requests
          {"WHERE start_time > '"+min_start_time+"'" if min_start_time else ""}
        """

    @staticmethod
    def get_table_storage_info() -> str:
        """Get detailed table storage information"""
        return """
               SELECT
                   OBJECT_SCHEMA_NAME(P.OBJECT_ID) AS SCHEMA_NAME,
                   OBJECT_NAME(P.OBJECT_ID) AS TABLE_NAME,
                   I.NAME AS INDEX_NAME,
                   I.TYPE_DESC AS INDEX_TYPE,
                   P.ROWS AS ROW_COUNT,
                   P.DATA_COMPRESSION_DESC AS COMPRESSION_TYPE,
                   AU.TOTAL_PAGES * 8.0 / 1024 AS TOTAL_SIZE_MB,
                   AU.USED_PAGES * 8.0 / 1024 AS USED_SIZE_MB,
                   CURRENT_TIMESTAMP AS EXTRACT_TS
               FROM SYS.PARTITIONS P
                        INNER JOIN SYS.INDEXES I ON P.OBJECT_ID = I.OBJECT_ID AND P.INDEX_ID = I.INDEX_ID
                        INNER JOIN SYS.ALLOCATION_UNITS AU ON P.PARTITION_ID = AU.CONTAINER_ID
               WHERE P.OBJECT_ID > 255
               ORDER BY AU.TOTAL_PAGES DESC
               """

    @staticmethod
    def get_query_performance_stats(days: int = 7) -> str:
        """Get query performance statistics"""
        return f"""
            SELECT
                qs.EXECUTION_COUNT,
                QS.TOTAL_ELAPSED_TIME / 1000000.0 AS TOTAL_ELAPSED_TIME_SEC,
                QS.TOTAL_WORKER_TIME / 1000000.0 AS TOTAL_WORKER_TIME_SEC,
                QS.TOTAL_LOGICAL_READS,
                QS.TOTAL_PHYSICAL_READS,
                QS.TOTAL_LOGICAL_WRITES,
                QS.LAST_EXECUTION_TIME,
                QS.CREATION_TIME,
                QS.LAST_ELAPSED_TIME / 1000000.0 AS LAST_ELAPSED_TIME_SEC,
                QS.LAST_WORKER_TIME / 1000000.0 AS LAST_WORKER_TIME_SEC,
                QS.LAST_LOGICAL_READS,
                QS.LAST_PHYSICAL_READS,
                QS.LAST_LOGICAL_WRITES,
                QS.TOTAL_ROWS,
                QS.LAST_ROWS,
                QS.MIN_ROWS,
                QS.MAX_ROWS,
                QS.STATEMENT_START_OFFSET,
                QS.STATEMENT_END_OFFSET,
                CURRENT_TIMESTAMP as EXTRACT_TS
            FROM SYS.DM_EXEC_QUERY_STATS QS
            WHERE QS.last_execution_time >= DATEADD(day, -{days}, GETDATE())
            ORDER BY QS.total_elapsed_time DESC
        """

    @staticmethod
    def list_query_stats(min_last_execution_time) -> str:
        """
        get the query stats
        source for below query:
        https://learn.microsoft.com/en-us/sql/relational-databases/system-dynamic-management-views/sys-dm-exec-query-stats-transact-sql?view=sql-server-ver16#b-returning-row-count-aggregates-for-a-query

        """
        return f"""
      SELECT
          SQL_HANDLE,
          PLAN_HANDLE,
          STATEMENT_START_OFFSET,
          STATEMENT_END_OFFSET,
          CREATION_TIME,
          LAST_EXECUTION_TIME,
          EXECUTION_COUNT,
          TOTAL_WORKER_TIME,
          LAST_WORKER_TIME,
          MIN_WORKER_TIME,
          MAX_WORKER_TIME,
          TOTAL_ELAPSED_TIME,
          LAST_ELAPSED_TIME,
          MIN_ELAPSED_TIME,
          MAX_ELAPSED_TIME,
          TOTAL_LOGICAL_READS,
          LAST_LOGICAL_READS,
          MIN_LOGICAL_READS,
          MAX_LOGICAL_READS,
          TOTAL_PHYSICAL_READS,
          LAST_PHYSICAL_READS,
          MIN_PHYSICAL_READS,
          MAX_PHYSICAL_READS,
          TOTAL_LOGICAL_WRITES,
          LAST_LOGICAL_WRITES,
          MIN_LOGICAL_WRITES,
          MAX_LOGICAL_WRITES,
          TOTAL_ROWS,
          LAST_ROWS,
          MIN_ROWS,
          MAX_ROWS,
          QUERY_HASH,
          QUERY_PLAN_HASH,
          SUBSTRING(
            ST.text,
            (QS.statement_start_offset / 2) + 1,
            (
              (
                CASE
                  statement_end_offset
                  WHEN -1 THEN DATALENGTH(ST.text)
                  ELSE QS.statement_end_offset
                END - QS.statement_start_offset
              ) / 2
            ) + 1
                ) AS statement_text
      FROM sys.dm_exec_query_stats AS QS
      CROSS APPLY sys.dm_exec_sql_text(QS.sql_handle) as ST
      {"WHERE QS.last_execution_time > '"+min_last_execution_time+"'" if min_last_execution_time else ""}"""

    @staticmethod
    def query_requests_history(min_end_time) -> str:
        # Serverless Request History
        return f"""SELECT
          STATUS,
          TRANSACTION_ID,
          DISTRIBUTED_STATEMENT_ID,
          QUERY_HASH,
          LOGIN_NAME,
          START_TIME,
          ERROR_CODE,
          REJECTED_ROWS_PATH,
          END_TIME,
          COMMAND,
          QUERY_TEXT,
          TOTAL_ELAPSED_TIME_MS,
          DATA_PROCESSED_MB,
          ERROR
        FROM sys.dm_exec_requests_history
        {"WHERE end_time > '"+min_end_time+"'" if min_end_time else ""}"""

    @staticmethod
    def data_processed(pool_name: str) -> str:
        return f"""
        SELECT
            DATA_PROCESSED_MB,
            TYPE,
            '{pool_name}' as POOL_NAME,
            CURRENT_TIMESTAMP AS EXTRACT_TS
            FROM SYS.DM_EXTERNAL_DATA_PROCESSED
        """



import logging

from pyspark.sql import DataFrame, SparkSession
from pyspark.sql.types import BooleanType, StringType, StructField, StructType
from sqlglot import Dialect, parse_one

from databricks.labs.lakebridge.reconcile.connectors.dialect_utils import DialectUtils
from databricks.labs.lakebridge.transpiler.sqlglot.dialect_utils import get_dialect
from databricks.labs.lakebridge.reconcile.recon_config import Schema, Table
from databricks.labs.lakebridge.reconcile.recon_output_config import SchemaMatchResult, SchemaReconcileOutput
from databricks.labs.lakebridge.transpiler.sqlglot.generator.databricks import Databricks

logger = logging.getLogger(__name__)


class SchemaCompare:
    def __init__(
        self,
        spark: SparkSession,
    ):
        self.spark = spark

    _schema_compare_output_schema: StructType = StructType(
        [
            StructField("source_column", StringType(), False),
            StructField("source_datatype", StringType(), False),
            StructField("databricks_column", StringType(), True),
            StructField("databricks_datatype", StringType(), True),
            StructField("is_valid", BooleanType(), False),
        ]
    )

    @classmethod
    def _build_master_schema(
        cls,
        source_schema: list[Schema],
        databricks_schema: list[Schema],
        table_conf: Table,
    ) -> list[SchemaMatchResult]:
        master_schema = SchemaCompare._select_columns(source_schema, table_conf)
        master_schema = SchemaCompare._drop_columns(master_schema, table_conf)

        target_column_map = table_conf.to_src_col_map or {}
        databricks_types_map = {c.ansi_normalized_column_name: c.data_type for c in databricks_schema}

        master_schema_match_res = [
            SchemaCompare.match_source_target_schemas(s, target_column_map, databricks_types_map) for s in master_schema
        ]

        return master_schema_match_res

    @staticmethod
    def _select_columns(master_schema: list[Schema], table_conf: Table):
        if table_conf.select_columns:
            return [schema for schema in master_schema if schema.column_name in table_conf.select_columns]
        return master_schema

    @staticmethod
    def _drop_columns(master_schema: list[Schema], table_conf: Table):
        if table_conf.drop_columns:
            return [sschema for sschema in master_schema if sschema.column_name not in table_conf.drop_columns]
        return master_schema

    @staticmethod
    def match_source_target_schemas(
        schema: Schema, target_column_map: dict, databricks_schema_map: dict
    ) -> SchemaMatchResult:
        databricks_column_name = target_column_map.get(
            schema.ansi_normalized_column_name, schema.ansi_normalized_column_name
        )
        databricks_datatype = databricks_schema_map.get(databricks_column_name, "Unknown")

        return SchemaMatchResult(
            source_column_normalized=schema.source_normalized_column_name,
            source_column_normalized_ansi=schema.ansi_normalized_column_name,
            source_datatype=schema.data_type,
            databricks_column=databricks_column_name,
            databricks_datatype=databricks_datatype,
        )

    def _create_output_dataframe(self, data: list[SchemaMatchResult], schema: StructType) -> DataFrame:
        """Return a user-friendly dataframe for schema compare result."""
        transformed = []
        for item in data:
            output = tuple(
                [
                    DialectUtils.unnormalize_identifier(item.source_column_normalized_ansi),
                    item.source_datatype,
                    DialectUtils.unnormalize_identifier(item.databricks_column),
                    item.databricks_datatype,
                    item.is_valid,
                ]
            )
            transformed.append(output)

        return self.spark.createDataFrame(transformed, schema)

    @classmethod
    def _table_schema_status(cls, schema_compare_maps: list[SchemaMatchResult]) -> bool:
        return bool(all(x.is_valid for x in schema_compare_maps))

    @classmethod
    def _validate_parsed_query(cls, source: Dialect, master: SchemaMatchResult) -> None:
        """
        Reconcile the schema of a single column by comparing the source column and the databricks column.

        1. This works by creating two SQL queries. both queries are a create table statement with a single column:
            * first query uses the source column name and datatype to simulate the source.
            * we escape the ansi name with sqlglot because normalize is not implemented for oracle and snowflake
            * second query uses the same column name as ansi and databricks datatype to simulate databricks.
            * we don't use the databricks column name as it may have been renamed.
            * renaming is checked in the previous step to retrieve the databricks column.
        2. Parse both queries using sqlglot and convert both to the other dialect
        3. Compare the converted queries with our original queries.
        4. If neither of the checks succeed, the column is marked as invalid

        :param source: source dialect e.g. TSQL, Oracle, Snowflake etc.
        :param master: source and target column names and datatypes computed by previous step.
        """
        target = get_dialect("databricks")
        source_column_normalized = cls._escape_source_column(source, target, master.source_column_normalized_ansi)
        source_query = f"create table dummy ({source_column_normalized} {master.source_datatype})"
        converted_source_query = cls._parse(source, target, source_query)
        databricks_query = f"create table dummy ({master.source_column_normalized_ansi} {master.databricks_datatype})"
        converted_databricks_query = cls._parse(target, source, databricks_query)
        parsed_source_check = converted_source_query.lower() == databricks_query.lower()
        parsed_databricks_check = source_query.lower() == converted_databricks_query.lower()
        logger.info(
            f"""
        Source query: {source_query}
        Converted source query: {converted_source_query}
        Databricks query: {databricks_query}
        Converted databricks query: {converted_databricks_query}
        Source equality check: {parsed_source_check}
        Databricks equality check: {parsed_databricks_check}
        """
        )

        if not parsed_source_check and not parsed_databricks_check:
            master.is_valid = False

    @classmethod
    def _parse(cls, source: Dialect, target: Dialect, source_query: str) -> str:
        return parse_one(source_query, read=source).sql(dialect=target).replace(", ", ",")

    @classmethod
    def _escape_source_column(cls, source: Dialect, target: Dialect, ansi_column: str) -> str:
        return parse_one(ansi_column, read=target).sql(dialect=source).replace(", ", ",")

    def compare(
        self,
        source_schema: list[Schema],
        databricks_schema: list[Schema],
        source: Dialect,
        table_conf: Table,
    ) -> SchemaReconcileOutput:
        """
        This method compares the source schema and the Databricks schema. It checks if the data types of the columns in the source schema
        match with the corresponding columns in the Databricks schema by parsing using remorph transpile.

        Returns:
            SchemaReconcileOutput: A dataclass object containing a boolean indicating the overall result of the comparison and a DataFrame with the comparison details.
        """
        master_schema = self._build_master_schema(source_schema, databricks_schema, table_conf)
        for master in master_schema:
            if not isinstance(source, Databricks):
                self._validate_parsed_query(source, master)
            elif master.source_datatype.lower() != master.databricks_datatype.lower():
                master.is_valid = False

        df = self._create_output_dataframe(master_schema, self._schema_compare_output_schema)
        final_result = self._table_schema_status(master_schema)
        return SchemaReconcileOutput(final_result, df)



import dataclasses

from databricks.labs.lakebridge.reconcile.connectors.data_source import DataSource
from databricks.labs.lakebridge.reconcile.recon_config import (
    Table,
    Aggregate,
    ColumnMapping,
    Transformation,
    ColumnThresholds,
)


class NormalizeReconConfigService:
    def __init__(self, source: DataSource, target: DataSource):
        self.source = source
        self.target = target

    def normalize_recon_table_config(self, table: Table) -> Table:
        normalized_table = dataclasses.replace(table)

        self._normalize_sampling(normalized_table)
        self._normalize_aggs(normalized_table)
        self._normalize_join_cols(normalized_table)
        self._normalize_select_cols(normalized_table)
        self._normalize_drop_cols(normalized_table)
        self._normalize_col_mappings(normalized_table)
        self._normalize_transformations(normalized_table)
        self._normalize_col_thresholds(normalized_table)
        self._normalize_jdbc_options(normalized_table)

        return normalized_table

    def _normalize_sampling(self, table: Table):
        if table.sampling_options:
            normalized_sampling = dataclasses.replace(table.sampling_options)
            normalized_sampling.stratified_columns = (
                [self.source.normalize_identifier(c).ansi_normalized for c in normalized_sampling.stratified_columns]
                if normalized_sampling.stratified_columns
                else None
            )
            table.sampling_options = normalized_sampling
        return table

    def _normalize_aggs(self, table: Table):
        normalized = [self._normalize_agg(a) for a in table.aggregates] if table.aggregates else None
        table.aggregates = normalized
        return table

    def _normalize_agg(self, agg: Aggregate) -> Aggregate:
        normalized = dataclasses.replace(agg)
        normalized.agg_columns = [self.source.normalize_identifier(c).ansi_normalized for c in normalized.agg_columns]
        normalized.group_by_columns = (
            [self.source.normalize_identifier(c).ansi_normalized for c in normalized.group_by_columns]
            if normalized.group_by_columns
            else None
        )
        return normalized

    def _normalize_join_cols(self, table: Table):
        table.join_columns = (
            [self.source.normalize_identifier(c).ansi_normalized for c in table.join_columns]
            if table.join_columns
            else None
        )
        return table

    def _normalize_select_cols(self, table: Table):
        table.select_columns = (
            [self.source.normalize_identifier(c).ansi_normalized for c in table.select_columns]
            if table.select_columns
            else None
        )
        return table

    def _normalize_drop_cols(self, table: Table):
        table.drop_columns = (
            [self.source.normalize_identifier(c).ansi_normalized for c in table.drop_columns]
            if table.drop_columns
            else None
        )
        return table

    def _normalize_col_mappings(self, table: Table):
        table.column_mapping = (
            [self._normalize_col_mapping(m) for m in table.column_mapping] if table.column_mapping else None
        )
        return table

    def _normalize_col_mapping(self, mapping: ColumnMapping):
        return ColumnMapping(
            source_name=self.source.normalize_identifier(mapping.source_name).ansi_normalized,
            target_name=self.target.normalize_identifier(mapping.target_name).ansi_normalized,
        )

    def _normalize_transformations(self, table: Table):
        table.transformations = (
            [self._normalize_transformation(t) for t in table.transformations] if table.transformations else None
        )
        return table

    def _normalize_transformation(self, transform: Transformation):
        """normalize user-configured transformations

        The user configures the table column and passes SQL code to transform the source table and target table.
        This is useful in scenarios when the data changes e.g. migrating `datetime`s. The SQL code is not normalized
        and it is the user responsibility to pass valid SQL respecting source database and target database.
        """
        normalized = dataclasses.replace(transform)
        normalized.column_name = self.source.normalize_identifier(transform.column_name).ansi_normalized
        return normalized

    def _normalize_col_thresholds(self, table: Table):
        table.column_thresholds = (
            [self._normalize_col_threshold(t) for t in table.column_thresholds] if table.column_thresholds else None
        )
        return table

    def _normalize_col_threshold(self, threshold: ColumnThresholds):
        normalized = dataclasses.replace(threshold)
        normalized.column_name = self.source.normalize_identifier(threshold.column_name).ansi_normalized
        return normalized

    def _normalize_jdbc_options(self, table: Table):
        if table.jdbc_reader_options:
            normalized = dataclasses.replace(table.jdbc_reader_options)
            normalized.partition_column = (
                self.source.normalize_identifier(normalized.partition_column).ansi_normalized
                if normalized.partition_column
                else None
            )
            table.jdbc_reader_options = normalized

        return table



import datetime
import logging
from pathlib import Path

from databricks.labs.lakebridge.intermediate.dag import DAG
from databricks.labs.lakebridge.intermediate.root_tables import RootTableAnalyzer
from databricks.labs.lakebridge.transpiler.sqlglot.sqlglot_engine import SqlglotEngine

logger = logging.getLogger(__name__)


def _generate_dot_file_contents(dag: DAG) -> str:
    _lineage_str = "flowchart TD\n"
    for node_name, node in dag.nodes.items():
        if node.parents:
            for parent in node.parents:
                _lineage_str += f"    {node_name.capitalize()} --> {parent.capitalize()}\n"
        else:
            # Include nodes without parents to ensure they appear in the diagram
            _lineage_str += f"    {node_name.capitalize()}\n"
    return _lineage_str


def lineage_generator(engine: SqlglotEngine, source_dialect: str, input_source: str, output_folder: str):
    input_sql_path = Path(input_source)
    output_folder = output_folder if output_folder.endswith('/') else output_folder + '/'

    msg = f"Processing for SQLs at this location: {input_sql_path}"
    logger.info(msg)
    root_table_analyzer = RootTableAnalyzer(engine, source_dialect, input_sql_path)
    generated_dag = root_table_analyzer.generate_lineage_dag()
    lineage_file_content = _generate_dot_file_contents(generated_dag)

    date_str = datetime.datetime.now().strftime("%d%m%y")

    output_filename = Path(f"{output_folder}lineage_{date_str}.dot")
    if output_filename.exists():
        logger.warning(f'The output file already exists and will be replaced: {output_filename}')
    logger.info(f"Attempting to write the lineage to {output_filename}")
    with output_filename.open('w', encoding='utf-8') as f:
        f.write(lineage_file_content)
    logger.info(f"Succeeded to write the lineage to {output_filename}")



import functools
import logging
from itertools import chain

from databricks.sdk import WorkspaceClient
from databricks.sdk.errors import NotFound
from databricks.sdk.service.catalog import (
    CatalogInfo,
    Privilege,
    SchemaInfo,
    SecurableType,
    VolumeInfo,
    VolumeType,
)

logger = logging.getLogger(__name__)


class CatalogOperations:
    def __init__(self, ws: WorkspaceClient):
        self._ws = ws

    def get_catalog(self, name: str) -> CatalogInfo | None:
        try:
            return self._ws.catalogs.get(name)
        except NotFound:
            return None

    def get_schema(self, catalog_name: str, schema_name: str) -> SchemaInfo | None:
        try:
            return self._ws.schemas.get(f"{catalog_name}.{schema_name}")
        except NotFound:
            return None

    def get_volume(self, catalog: str, schema: str, name: str) -> VolumeInfo | None:
        try:
            return self._ws.volumes.read(f"{catalog}.{schema}.{name}")
        except NotFound:
            return None

    def create_catalog(self, name: str) -> CatalogInfo:
        logger.debug(f"Creating catalog `{name}`.")
        catalog_info = self._ws.catalogs.create(name)
        logger.info(f"Created catalog `{name}`.")
        return catalog_info

    def create_schema(self, schema_name: str, catalog_name: str) -> SchemaInfo:
        logger.debug(f"Creating schema `{schema_name}` in catalog `{catalog_name}`.")
        schema_info = self._ws.schemas.create(schema_name, catalog_name)
        logger.info(f"Created schema `{schema_name}` in catalog `{catalog_name}`.")
        return schema_info

    def create_volume(
        self,
        catalog: str,
        schema: str,
        name: str,
        volume_type: VolumeType = VolumeType.MANAGED,
    ) -> VolumeInfo:
        logger.debug(f"Creating volume `{name}` in catalog `{catalog}` and schema `{schema}`")
        volume_info = self._ws.volumes.create(catalog, schema, name, volume_type)
        logger.info(f"Created volume `{name}` in catalog `{catalog}` and schema `{schema}`")
        return volume_info

    def has_catalog_access(
        self,
        catalog: CatalogInfo,
        user_name: str,
        privilege_sets: tuple[set[Privilege], ...],
    ) -> bool:
        """
        Check if a user has access to a catalog based on ownership or a set of privileges.
        :param catalog: A catalog to check access for.
        :param user_name: Username to check.
        :param privilege_sets: A tuple of sets, where each set contains Privilege objects.
                               The function checks if the user has any of these sets of privileges. For example:
                               ({Privilege.ALL_PRIVILEGES}, {Privilege.USE_CATALOG, Privilege.APPLY_TAG})
                               In this case, the user would need either ALL_PRIVILEGES,
                               or both USE_CATALOG and APPLY_TAG.
        """
        if user_name == catalog.owner:
            return True

        return any(
            self.has_privileges(user_name, SecurableType.CATALOG, catalog.name, privilege_set)
            for privilege_set in privilege_sets
        )

    def has_schema_access(
        self,
        schema: SchemaInfo,
        user_name: str,
        privilege_sets: tuple[set[Privilege], ...],
    ) -> bool:
        """
        Check if a user has access to a schema based on ownership or a set of privileges.
        :param schema: A schema to check access for.
        :param user_name: Username to check.
        :param privilege_sets: The function checks if the user has any of these sets of privileges. For example:
                               ({Privilege.ALL_PRIVILEGES}, {Privilege.USE_SCHEMA, Privilege.CREATE_TABLE})
                               In this case, the user would need either ALL_PRIVILEGES,
                               or both USE_SCHEMA and CREATE_TABLE.
        """
        if user_name == schema.owner:
            return True

        return any(
            self.has_privileges(user_name, SecurableType.SCHEMA, schema.full_name, privilege_set)
            for privilege_set in privilege_sets
        )

    def has_volume_access(
        self,
        volume: VolumeInfo,
        user_name: str,
        privilege_sets: tuple[set[Privilege], ...],
    ) -> bool:
        """
        Check if a user has access to a volume based on ownership or a set of privileges.
        :param volume: A volume to check access for.
        :param user_name: Username to check.
        :param privilege_sets: The function checks if the user has any of these sets of privileges. For example:
                               ({Privilege.ALL_PRIVILEGES}, {Privilege.READ_VOLUME, Privilege.WRITE_VOLUME})
                               In this case, the user would need either ALL_PRIVILEGES,
                               or both READ_VOLUME and WRITE_VOLUME.
        """
        if user_name == volume.owner:
            return True

        return any(
            self.has_privileges(user_name, SecurableType.VOLUME, volume.full_name, privilege_set)
            for privilege_set in privilege_sets
        )

    def has_privileges(
        self,
        user: str | None,
        securable_type: SecurableType,
        full_name: str | None,
        privileges: set[Privilege],
    ) -> bool:
        """
        Check if a user has a set of privileges for a securable object.
        """
        assert user, "User must be provided"
        assert full_name, "Full name must be provided"
        user_privileges = self._get_user_privileges(user, securable_type, full_name)
        result = privileges.issubset(user_privileges)
        if not result:
            logger.debug(f"User {user} doesn't have privilege set {privileges} for {securable_type} {full_name}")
        return result

    @functools.lru_cache(maxsize=1024)
    def _get_user_privileges(self, user: str, securable_type: SecurableType, full_name: str) -> set[Privilege]:
        permissions = self._ws.grants.get_effective(securable_type.value, full_name, principal=user)
        if not permissions or not permissions.privilege_assignments:
            return set()
        return {
            p.privilege
            for p in chain.from_iterable(
                privilege.privileges for privilege in permissions.privilege_assignments if privilege.privileges
            )
            if p.privilege
        }



import logging
from abc import ABC, abstractmethod

from pyspark.sql import DataFrame

from databricks.labs.lakebridge.reconcile.connectors.dialect_utils import DialectUtils
from databricks.labs.lakebridge.reconcile.connectors.models import NormalizedIdentifier
from databricks.labs.lakebridge.reconcile.exception import DataSourceRuntimeException
from databricks.labs.lakebridge.reconcile.recon_config import JdbcReaderOptions, Schema

logger = logging.getLogger(__name__)


class DataSource(ABC):

    @abstractmethod
    def read_data(
        self,
        catalog: str | None,
        schema: str,
        table: str,
        query: str,
        options: JdbcReaderOptions | None,
    ) -> DataFrame:
        return NotImplemented

    @abstractmethod
    def get_schema(
        self,
        catalog: str | None,
        schema: str,
        table: str,
        normalize: bool = True,
    ) -> list[Schema]:
        return NotImplemented

    @abstractmethod
    def normalize_identifier(self, identifier: str) -> NormalizedIdentifier:
        pass

    @classmethod
    def log_and_throw_exception(cls, exception: Exception, fetch_type: str, query: str):
        error_msg = f"Runtime exception occurred while fetching {fetch_type} using {query} : {exception}"
        logger.warning(error_msg)
        raise DataSourceRuntimeException(error_msg) from exception

    def _map_meta_column(self, meta_column, normalize: bool) -> Schema:
        """Create a normalized Schema DTO from the database metadata

        Used in the implementations of get_schema to build a Schema DTO from the `INFORMATION_SCHEMA` query result.
        The returned Schema is normalized in case the database is having columns with special characters and standardize
        """
        name = meta_column.column_name.lower()
        dtype = meta_column.data_type.strip().lower()
        if normalize:
            normalized = self.normalize_identifier(name)
            return Schema(normalized.ansi_normalized, dtype, normalized.ansi_normalized, normalized.source_normalized)

        return Schema(name, dtype, name, name)


class MockDataSource(DataSource):

    def __init__(
        self,
        dataframe_repository: dict[tuple[str, str, str], DataFrame],
        schema_repository: dict[tuple[str, str, str], list[Schema]],
        exception: Exception = RuntimeError("Mock Exception"),
        delimiter: str = "`",
    ):
        self._dataframe_repository: dict[tuple[str, str, str], DataFrame] = dataframe_repository
        self._schema_repository: dict[tuple[str, str, str], list[Schema]] = schema_repository
        self._exception = exception
        self._delimiter = delimiter

    def read_data(
        self,
        catalog: str | None,
        schema: str,
        table: str,
        query: str,
        options: JdbcReaderOptions | None,
    ) -> DataFrame:
        catalog_str = catalog if catalog else ""
        mock_df = self._dataframe_repository.get((catalog_str, schema, query))
        if not mock_df:
            return self.log_and_throw_exception(self._exception, "data", f"({catalog}, {schema}, {query})")
        return mock_df

    def get_schema(self, catalog: str | None, schema: str, table: str, normalize: bool = True) -> list[Schema]:
        catalog_str = catalog if catalog else ""
        mock_schema = self._schema_repository.get((catalog_str, schema, table))
        if not mock_schema:
            return self.log_and_throw_exception(self._exception, "schema", f"({catalog}, {schema}, {table})")
        return mock_schema

    def normalize_identifier(self, identifier: str) -> NormalizedIdentifier:
        return DialectUtils.normalize_identifier(identifier, self._delimiter, self._delimiter)



import logging

from sqlglot import Dialect
from sqlglot import expressions as exp

from databricks.labs.lakebridge.reconcile.query_builder.expression_generator import build_column, build_literal
from databricks.labs.lakebridge.reconcile.recon_config import Table

logger = logging.getLogger(__name__)


class CountQueryBuilder:

    def __init__(
        self,
        table_conf: Table,
        layer: str,
        engine: Dialect,
    ):
        self._table_conf = table_conf
        self._layer = layer
        self._engine = engine

    def build_query(self):
        select_clause = build_column(this=exp.Count(this=build_literal(this="1", is_string=False)), alias="count")
        count_query = (
            exp.select(select_clause)
            .from_(":tbl")
            .where(self._table_conf.get_filter(self._layer))
            .sql(dialect=self._engine)
        )
        logger.info(f"Record Count Query for {self._layer}: {count_query}")
        return count_query



def refactor_hexadecimal_chars(input_string: str) -> str:
    """
    Updates the HexaDecimal characters ( \x1b[\\d+m ) in the given string as below.
    :param input_string: String with HexaDecimal characters. ex: ( \x1b[4mWHERE\x1b[0m )
    :return: String with HexaDecimal characters refactored to arrows. ex: ( --> WHERE <--)
    """
    output_string = input_string
    highlight = {"\x1b[4m": "--> ", "\x1b[0m": " <--"}
    for key, value in highlight.items():
        output_string = output_string.replace(key, value)
    return output_string


def format_error_message(error_type: str, error_message: Exception, error_sql: str) -> str:
    """
    Formats the error message with the error SQL.
    :param error_type: Error Type
    :param error_message: Error message
    :param error_sql: Error SQL
    :return: Formatted error message
    """
    error_str = (
        f"------------------------ {error_type} Start:------------------------\n"
        f"/*\n{str(error_message)}\n*/\n\n"
        f"/*\nOriginal Query:\n\n{str(error_sql)}\n*/\n"
        f"------------------------- {error_type} End:-------------------------"
    ).strip()
    return error_str



import logging
from importlib.resources import files

from databricks.labs.blueprint.installation import Installation
from databricks.labs.blueprint.installer import InstallState
from databricks.labs.blueprint.wheels import ProductInfo
from databricks.labs.blueprint.wheels import find_project_root
from databricks.sdk import WorkspaceClient
from databricks.sdk.errors import InvalidParameterValue, NotFound

import databricks.labs.lakebridge.resources
from databricks.labs.lakebridge.config import ReconcileConfig
from databricks.labs.lakebridge.deployment.dashboard import DashboardDeployment
from databricks.labs.lakebridge.deployment.job import JobDeployment
from databricks.labs.lakebridge.deployment.table import TableDeployment

logger = logging.getLogger(__name__)

_RECON_PREFIX = "Reconciliation"
RECON_JOB_NAME = f"{_RECON_PREFIX} Runner"


class ReconDeployment:
    def __init__(
        self,
        ws: WorkspaceClient,
        installation: Installation,
        install_state: InstallState,
        product_info: ProductInfo,
        table_deployer: TableDeployment,
        job_deployer: JobDeployment,
        dashboard_deployer: DashboardDeployment,
    ):
        self._ws = ws
        self._installation = installation
        self._install_state = install_state
        self._product_info = product_info
        self._table_deployer = table_deployer
        self._job_deployer = job_deployer
        self._dashboard_deployer = dashboard_deployer

    def install(self, recon_config: ReconcileConfig | None, wheel_path: str):
        if not recon_config:
            logger.warning("Recon Config is empty.")
            return
        logger.info("Installing reconcile components.")
        self._deploy_tables(recon_config)
        self._deploy_dashboards(recon_config)
        self._deploy_jobs(recon_config, wheel_path)
        self._install_state.save()
        logger.info("Installation of reconcile components completed successfully.")

    def uninstall(self, recon_config: ReconcileConfig | None):
        if not recon_config:
            return
        logger.info("Uninstalling reconcile components.")
        self._remove_dashboards()
        self._remove_jobs()
        logging.info(
            f"Won't remove reconcile metadata schema `{recon_config.metadata_config.schema}` "
            f"from catalog `{recon_config.metadata_config.catalog}`. Please remove it and the tables inside manually."
        )
        logging.info(
            f"Won't remove configured reconcile secret scope `{recon_config.secret_scope}`. "
            f"Please remove it manually."
        )

    def _deploy_tables(self, recon_config: ReconcileConfig):
        logger.info("Deploying reconciliation metadata tables.")
        catalog = recon_config.metadata_config.catalog
        schema = recon_config.metadata_config.schema
        resources = files(databricks.labs.lakebridge.resources)
        query_dir = resources.joinpath("reconcile/queries/installation")

        sqls_to_deploy = [
            "main.sql",
            "metrics.sql",
            "details.sql",
            "aggregate_metrics.sql",
            "aggregate_details.sql",
            "aggregate_rules.sql",
        ]

        for sql_file in sqls_to_deploy:
            table_sql_file = query_dir.joinpath(sql_file)
            self._table_deployer.deploy_table_from_ddl_file(catalog, schema, sql_file.strip(".sql"), table_sql_file)

    def _deploy_dashboards(self, recon_config: ReconcileConfig):
        logger.info("Deploying reconciliation dashboards.")
        dashboard_base_dir = (
            find_project_root(__file__) / "src/databricks/labs/lakebridge/resources/reconcile/dashboards"
        )
        self._dashboard_deployer.deploy(dashboard_base_dir, recon_config.metadata_config)

    def _get_dashboards(self) -> list[tuple[str, str]]:
        return list(self._install_state.dashboards.items())

    def _remove_dashboards(self):
        logger.info("Removing reconciliation dashboards.")
        for dashboard_ref, dashboard_id in self._get_dashboards():
            try:
                logger.info(f"Removing dashboard with id={dashboard_id}.")
                del self._install_state.dashboards[dashboard_ref]
                self._ws.lakeview.trash(dashboard_id)
            except (InvalidParameterValue, NotFound):
                logger.warning(f"Dashboard with id={dashboard_id} doesn't exist anymore for some reason.")
                continue

    def _deploy_jobs(self, recon_config: ReconcileConfig, lakebridge_wheel_path: str):
        logger.info("Deploying reconciliation jobs.")
        self._job_deployer.deploy_recon_job(RECON_JOB_NAME, recon_config, lakebridge_wheel_path)
        for job_name, job_id in self._get_deprecated_jobs():
            try:
                logger.info(f"Removing job_id={job_id}, as it is no longer needed.")
                del self._install_state.jobs[job_name]
                self._ws.jobs.delete(job_id)
            except (InvalidParameterValue, NotFound):
                logger.warning(f"{job_name} doesn't exist anymore for some reason.")
                continue

    def _get_jobs(self) -> list[tuple[str, int]]:
        return [
            (job_name, int(job_id))
            for job_name, job_id in self._install_state.jobs.items()
            if job_name.startswith(_RECON_PREFIX)
        ]

    def _get_deprecated_jobs(self) -> list[tuple[str, int]]:
        return [
            (job_name, int(job_id))
            for job_name, job_id in self._install_state.jobs.items()
            if job_name.startswith(_RECON_PREFIX) and job_name != RECON_JOB_NAME
        ]

    def _remove_jobs(self):
        logger.info("Removing Reconciliation Jobs.")
        for job_name, job_id in self._get_jobs():
            try:
                logger.info(f"Removing job {job_name} with job_id={job_id}.")
                del self._install_state.jobs[job_name]
                self._ws.jobs.delete(int(job_id))
            except (InvalidParameterValue, NotFound):
                logger.warning(f"{job_name} doesn't exist anymore for some reason.")
                continue



import json
import logging
import os
import sys
import venv
import tempfile
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from subprocess import CalledProcessError, PIPE, Popen, STDOUT, run

import duckdb
import yaml

from databricks.labs.blueprint.paths import read_text
from databricks.labs.lakebridge.assessments.profiler_config import PipelineConfig, Step
from databricks.labs.lakebridge.connections.credential_manager import cred_file
from databricks.labs.lakebridge.connections.database_manager import DatabaseManager, FetchResult

logger = logging.getLogger(__name__)

DB_NAME = "profiler_extract.db"


class StepExecutionStatus(str, Enum):
    COMPLETE = "COMPLETE"
    ERROR = "ERROR"
    SKIPPED = "SKIPPED"


@dataclass
class StepExecutionResult:
    step_name: str
    status: StepExecutionStatus
    error_message: str | None = None


class PipelineClass:
    def __init__(self, config: PipelineConfig, executor: DatabaseManager | None):
        self.config = config
        self.executor = executor
        self._db_path_prefix = Path(config.extract_folder).expanduser()
        self._create_dir(self._db_path_prefix)
        self._db_path = str(self._db_path_prefix / DB_NAME)

    def execute(self) -> list[StepExecutionResult]:
        logging.info(f"Pipeline initialized with config: {self.config.name}, version: {self.config.version}")
        execution_results: list[StepExecutionResult] = []

        for step in self.config.steps:
            result = self._process_step(step)
            execution_results.append(result)
            self._log_step_result(result)

            # Fail immediately if DDL step failed
            if step.type == "ddl" and result.status == StepExecutionStatus.ERROR:
                error_msg = f"Pipeline execution failed due to error in DDL step: {result.step_name}"
                if result.error_message:
                    error_msg += f" - {result.error_message}"
                logger.error(error_msg)
                raise RuntimeError(error_msg)

        # Check if any non-DDL steps failed
        failed_steps = [r for r in execution_results if r.status == StepExecutionStatus.ERROR]
        if failed_steps:
            error_msg = (
                f"Pipeline execution failed due to errors in steps: {', '.join(r.step_name for r in failed_steps)}"
            )
            logger.error(error_msg)
            raise RuntimeError(error_msg)

        return execution_results

    def _process_step(self, step: Step) -> StepExecutionResult:
        logger.info(f"Executing step: {step.name}")

        if step.flag != "active":
            logging.info(f"Skipping step: {step.name} as it is not active")
            return StepExecutionResult(step_name=step.name, status=StepExecutionStatus.SKIPPED)

        try:
            # Execute based on step type
            match step.type:
                case "sql":
                    self._execute_sql_step(step)
                case "ddl":
                    self._execute_ddl_step(step)
                case "python":
                    self._execute_python_step(step)
                case _:
                    raise RuntimeError(f"Unsupported step type: {step.type}")

            return StepExecutionResult(step_name=step.name, status=StepExecutionStatus.COMPLETE)
        except RuntimeError as e:
            return StepExecutionResult(step_name=step.name, status=StepExecutionStatus.ERROR, error_message=str(e))

    def _log_step_result(self, result: StepExecutionResult):
        match result.status:
            case StepExecutionStatus.ERROR:
                logger.error(f"Step {result.step_name} failed with error: {result.error_message}")
            case StepExecutionStatus.SKIPPED:
                logger.info(f"Step {result.step_name} was skipped.")
            case StepExecutionStatus.COMPLETE:
                logger.info(f"Step {result.step_name} has completed successfully.")

    def _execute_sql_step(self, step: Step):
        logging.debug(f"Reading query from file: {step.extract_source}")
        query = read_text(Path(step.extract_source))

        if self.executor is None:
            logging.error("DatabaseManager executor is not set.")
            raise RuntimeError("DatabaseManager executor is not set.")

        # Execute the query using the database manager
        logging.info(f"Executing query: {query}")
        try:
            result = self.executor.fetch(query)

            # Save the result to duckdb
            self._save_to_db(result, step.name, str(step.mode))
        except Exception as e:
            logging.error(f"SQL execution failed: {str(e)}")
            raise RuntimeError(f"SQL execution failed: {str(e)}") from e

    def _execute_ddl_step(self, step: Step):
        logging.debug(f"Reading DDL from file: {step.extract_source}")
        ddl = read_text(Path(step.extract_source)).strip()

        logging.info(f"Executing DDL for table '{step.name}'")

        try:
            # TODO: Handle schema evolution
            # Current implementation just checks for table existence;
            # mode logic becomes irrelevant for ddl step.
            with duckdb.connect(self._db_path) as conn:
                conn.begin()
                if not self._table_exists(conn, step.name):
                    conn.execute(ddl)
                    conn.commit()
                    logging.debug(f"Created new table '{step.name}'")
                else:
                    logging.debug(f"Table '{step.name}' already exists, skipping DDL execution")
        except Exception as e:
            logging.error(f"DDL execution failed: {str(e)}")
            raise RuntimeError(f"DDL execution failed: {str(e)}") from e

    def _execute_python_step(self, step: Step):

        logging.debug(f"Executing Python script: {step.extract_source}")
        credential_config = str(cred_file("lakebridge"))
        venv_path_prefix = Path.home() / ".databricks" / "labs" / "lakebridge_profilers"
        os.makedirs(venv_path_prefix, exist_ok=True)

        # Create a temporary directory for the virtual environment
        # TODO Windows has strict checks on for temp venv cleanup, so will ignore cleanup errors and have it cleaned up later
        with tempfile.TemporaryDirectory(dir=venv_path_prefix, ignore_cleanup_errors=True) as temp_dir:
            venv_dir = Path(temp_dir) / "venv"
            venv_exec_cmd = self._create_venv(venv_dir)

            # Define the paths to the virtual environment's Python and pip executables
            if sys.platform == "win32":
                venv_python = (venv_dir / "Scripts" / "python.exe").resolve()
                venv_pip = (venv_dir / "Scripts" / "pip.exe").resolve()
            else:
                venv_python = (venv_dir / "bin" / "python").resolve()
                venv_pip = (venv_dir / "bin" / "pip").resolve()

            # Log resolved paths
            logger.info(f"Resolved venv_python: {venv_python}")
            logger.info(f"Resolved venv_pip: {venv_pip}")

            logger.info(f"Creating a virtual environment for Python script execution: {venv_dir} for step: {step.name}")
            if step.dependencies:
                self._install_dependencies(venv_exec_cmd, step.dependencies)

            self._run_python_script(venv_exec_cmd, step.extract_source, self._db_path, credential_config)

    @staticmethod
    def _install_dependencies(venv_exec_cmd, dependencies):
        logging.info(f"Installing dependencies: {', '.join(dependencies)}")
        try:
            logging.debug("Upgrading local pip")
            is_debug = logging.getLogger(__name__).isEnabledFor(logging.DEBUG)
            run(
                [
                    venv_exec_cmd,
                    "-m",
                    "pip",
                    "install",
                    "--upgrade",
                    "pip",
                    "--require-virtualenv",
                    "--no-input",
                    "--disable-pip-version-check",
                ],
                check=True,
                capture_output=not is_debug,
                text=True,
            )

            run(
                [
                    venv_exec_cmd,
                    "-m",
                    "pip",
                    "install",
                    *dependencies,
                    "--require-virtualenv",
                    "--no-input",
                    "--disable-pip-version-check",
                ],
                check=True,
                capture_output=not is_debug,
                text=True,
            )
        except CalledProcessError as e:
            # Log detailed output at debug level for troubleshooting
            logging.debug(
                f"Failed to install dependencies (exit code {e.returncode})\n" f"stdout: {e.stdout}\nstderr: {e.stderr}"
            )
            logging.error(f"Failed to install dependencies: {e.stderr}")
            raise RuntimeError(f"Failed to install dependencies: {e.stderr}") from e

    @staticmethod
    def _run_python_script(venv_exec_cmd, script_path, db_path, credential_config):
        output_lines = []
        try:
            with Popen(
                [
                    venv_exec_cmd,
                    str(script_path),
                    "--db-path",
                    db_path,
                    "--credential-config-path",
                    credential_config,
                ],
                stdout=PIPE,
                stderr=STDOUT,
                text=True,
                bufsize=1,
            ) as process:
                if process.stdout is not None:
                    for line in process.stdout:
                        logger.info(line.rstrip())
                        output_lines.append(line)
                process.wait()
        except Exception as e:
            logging.error(f"Python script failed: {str(e)}")
            raise RuntimeError(f"Script execution failed: {str(e)}") from e

        if output_lines:
            try:
                output = json.loads(output_lines[-1])
            except json.JSONDecodeError:
                logging.info("Could not parse script output as JSON.")
                output = {
                    "status": "error",
                    "message": "Could not parse script output as JSON, manually validate the logs.",
                }

            if output.get("status") == "success":
                logging.info(f"Python script completed: {output['message']}")
            else:
                raise RuntimeError(f"Script reported error: {output.get('message', 'Unknown error')}")

        if process.returncode != 0:
            raise RuntimeError(f"Script execution failed with exit code {process.returncode}")

    def _save_to_db(self, result: FetchResult, step_name: str, mode: str):
        # Check row count and log appropriately and skip data insertion if 0 rows
        if not result.rows:
            logging.warning(
                f"Query for step '{step_name}' returned 0 rows. Skipping table creation and data insertion."
            )
            return

        row_count = len(result.rows)
        logging.info(f"Query for step '{step_name}' returned {row_count} rows.")

        with duckdb.connect(self._db_path) as conn:
            # Note: step_name is validated to be SQL-safe by Step.__post_init__
            table_exists = self._table_exists(conn, step_name)
            conn.begin()
            if table_exists and mode == 'overwrite':
                # Table exists and overwrite mode: Truncate then insert within a transaction to preserve existing DDL schema
                _result_frame = result.to_df()
                # Note: step_name is validated to be SQL-safe by Step.__post_init__
                logging.debug(f"Overwriting existing table '{step_name}'")
                conn.execute(f"TRUNCATE {step_name}")
                conn.execute(f"INSERT INTO {step_name} SELECT * FROM _result_frame")
            else:
                if table_exists:
                    # Table exists and append mode: insert into existing table (DuckDB handles type conversion)
                    _result_frame = result.to_df()
                    # Note: step_name is validated to be SQL-safe by Step.__post_init__
                    statement = f"INSERT INTO {step_name} SELECT * FROM _result_frame"
                    logging.debug(f"Appending to existing table '{step_name}'")
                else:
                    # Table doesn't exist: create table with native types from query result
                    # Use DDL steps for explicit type control when needed
                    _result_frame = result.to_df()
                    # Note: step_name is validated to be SQL-safe by Step.__post_init__
                    statement = f"CREATE TABLE {step_name} AS SELECT * FROM _result_frame"
                    logging.debug(f"Creating new table '{step_name}' with native types")

                logging.debug(f"Executing: {statement}")
                conn.execute(statement)

            # Explicit commit before context exit
            conn.commit()
            logging.info(f"Successfully processed {row_count} rows for table '{step_name}'.")

    @staticmethod
    def _table_exists(conn: duckdb.DuckDBPyConnection, table_name: str) -> bool:
        result = conn.execute(
            "SELECT COUNT(*) FROM information_schema.tables WHERE table_name = ?", [table_name]
        ).fetchone()
        return result[0] > 0 if result else False

    @staticmethod
    def _create_dir(dir_path: Path):
        if not Path(dir_path).exists():
            dir_path.mkdir(parents=True, exist_ok=True)

    @staticmethod
    def load_config_from_yaml(file_path: str | Path) -> PipelineConfig:
        with open(file_path, 'r', encoding='utf-8') as file:
            data = yaml.safe_load(file)
        steps = [Step(**step) for step in data['steps']]
        return PipelineConfig(
            name=data['name'], version=data['version'], extract_folder=data['extract_folder'], steps=steps
        )

    @staticmethod
    def _create_venv(install_path: Path) -> str:
        venv_path = install_path
        # Sadly, some platform-specific variations need to be dealt with:
        #   - Windows venvs do not use symlinks, but rather copies, when populating the venv.
        #   - The library path is different.
        use_symlinks = sys.platform != "win32"

        builder = venv.EnvBuilder(with_pip=True, symlinks=use_symlinks)
        builder.create(venv_path)
        context = builder.ensure_directories(venv_path)
        logger.debug(f"Created virtual environment with context: {context}")
        return context.env_exec_cmd



import logging

from databricks.labs.blueprint.installer import InstallState
from databricks.sdk import WorkspaceClient
from databricks.sdk.service._internal import Wait
from databricks.sdk.service.jobs import Run

from databricks.labs.lakebridge.deployment.recon import RECON_JOB_NAME
from databricks.labs.lakebridge.reconcile.recon_config import RECONCILE_OPERATION_NAME

logger = logging.getLogger(__name__)

_RECON_DOCS_URL = "https://databrickslabs.github.io/lakebridge/docs/reconcile/"


class ReconcileRunner:
    def __init__(
        self,
        ws: WorkspaceClient,
        install_state: InstallState,
    ):
        self._ws = ws
        self._install_state = install_state

    def run(self, operation_name: str = RECONCILE_OPERATION_NAME) -> tuple[Wait[Run], str]:
        job_id = self._get_recon_job_id()
        logger.info(f"Triggering the reconcile job with job_id: `{job_id}`")
        wait = self._ws.jobs.run_now(job_id, job_parameters={"operation_name": operation_name})
        if not wait.run_id:
            raise SystemExit(f"Job {job_id} execution failed. Please check the job logs for more details.")

        job_run_url = f"{self._ws.config.host}/jobs/{job_id}/runs/{wait.run_id}"
        logger.info(
            f"'{operation_name.upper()}' job started. Please check the job_url `{job_run_url}` for the current status."
        )
        return wait, job_run_url

    def _get_recon_job_id(self) -> int:
        if RECON_JOB_NAME in self._install_state.jobs:
            logger.debug("Reconcile job id found in the install state.")
            return int(self._install_state.jobs[RECON_JOB_NAME])
        raise SystemExit("Reconcile Job ID not found. Please try reinstalling.")



import logging
import os
import sys

from databricks.connect import DatabricksSession
from databricks.labs.blueprint.installation import Installation
from databricks.sdk import WorkspaceClient

from databricks.labs.lakebridge import initialize_logging
from databricks.labs.lakebridge.config import ReconcileConfig, TableRecon
from databricks.labs.lakebridge.reconcile.recon_config import AGG_RECONCILE_OPERATION_NAME, RECONCILE_OPERATION_NAME
from databricks.labs.lakebridge.reconcile.trigger_recon_aggregate_service import TriggerReconAggregateService
from databricks.labs.lakebridge.reconcile.trigger_recon_service import TriggerReconService

logger = logging.getLogger(__name__)


def main(*argv: str) -> None:
    """Lakeview Jobs task entry point: reconcile"""
    initialize_logging()

    logger.debug(f"Arguments received: {argv}")
    w = WorkspaceClient()
    installation: Installation | None = None
    operation_name: str | None = None
    # sys.arg is used when running the script as an entry point which is how we trigger the job.
    args = argv[1:] if argv else tuple(sys.argv[1:])
    match args:
        case [operation_name, install_folder] if operation_name in {
            RECONCILE_OPERATION_NAME,
            AGG_RECONCILE_OPERATION_NAME,
        }:
            installation = Installation(w, "lakebridge", install_folder=install_folder)
        case [operation_name] if operation_name in {
            RECONCILE_OPERATION_NAME,
            AGG_RECONCILE_OPERATION_NAME,
        }:
            installation = Installation.assume_user_home(w, "lakebridge")
        case _:
            raise ValueError(
                f"Invalid arguments: {args}. Expected [operation_name, install_folder] "
                f"where operation_name is one of: {RECONCILE_OPERATION_NAME!r}, {AGG_RECONCILE_OPERATION_NAME!r}."
            )

    reconcile_config = installation.load(ReconcileConfig)

    catalog_or_schema = (
        reconcile_config.database_config.source_catalog
        if reconcile_config.database_config.source_catalog
        else reconcile_config.database_config.source_schema
    )
    filename = f"recon_config_{reconcile_config.data_source}_{catalog_or_schema}_{reconcile_config.report_type}.json"

    logger.info(f"Loading {filename} from Databricks Workspace...")

    table_recon = installation.load(type_ref=TableRecon, filename=filename)

    if operation_name == AGG_RECONCILE_OPERATION_NAME:
        return _trigger_reconcile_aggregates(w, table_recon, reconcile_config)

    return _trigger_recon(w, table_recon, reconcile_config)


def _trigger_recon(
    w: WorkspaceClient,
    table_recon: TableRecon,
    reconcile_config: ReconcileConfig,
):
    recon_output = TriggerReconService.trigger_recon(
        ws=w,
        spark=DatabricksSession.builder.getOrCreate(),
        table_recon=table_recon,
        reconcile_config=reconcile_config,
    )
    logger.info(f"Output: {recon_output}")


def _trigger_reconcile_aggregates(
    ws: WorkspaceClient,
    table_recon: TableRecon,
    reconcile_config: ReconcileConfig,
):
    """
    Triggers the reconciliation process for aggregated data  between source and target tables.
    Supported Aggregate functions: MIN, MAX, COUNT, SUM, AVG, MEAN, MODE, PERCENTILE, STDDEV, VARIANCE, MEDIAN

    This function attempts to reconcile aggregate data based on the configurations provided. It logs the outcome
    of the reconciliation process, including any errors encountered during execution.

    Parameters:
    - ws (WorkspaceClient): The workspace client used to interact with Databricks workspaces.
    - table_recon (TableRecon): Configuration for the table reconciliation process, including source and target details.
    - reconcile_config (ReconcileConfig): General configuration for the reconciliation process,
                                                                    including database and table settings.

    Raises:
    - ReconciliationException: If an error occurs during the reconciliation process, it is caught and re-raised
      after logging the error details.
    """
    reconcile_config.report_type = "aggregate"
    recon_output = TriggerReconAggregateService.trigger_recon_aggregates(
        ws=ws,
        spark=DatabricksSession.builder.getOrCreate(),
        table_recon=table_recon,
        reconcile_config=reconcile_config,
    )
    logger.info(f"Output: {recon_output}")


if __name__ == "__main__":
    if "DATABRICKS_RUNTIME_VERSION" not in os.environ:
        raise SystemExit("Only intended to run in Databricks Runtime")
    main(*sys.argv)



import dataclasses
import logging
import os
import sys
import webbrowser
from collections.abc import Callable, Sequence, Set
from pathlib import Path
from typing import Any

from databricks.labs.blueprint.entrypoint import get_logger
from databricks.labs.blueprint.installation import Installation, JsonValue, SerdeError
from databricks.labs.blueprint.installer import InstallState
from databricks.labs.blueprint.tui import Prompts
from databricks.labs.blueprint.wheels import ProductInfo
from databricks.sdk import WorkspaceClient
from databricks.sdk.core import with_user_agent_extra
from databricks.sdk.errors import NotFound, PermissionDenied

from databricks.labs.lakebridge import initialize_logging
from databricks.labs.lakebridge.__about__ import __version__
from databricks.labs.lakebridge.cli import lakebridge
from databricks.labs.lakebridge.config import (
    DatabaseConfig,
    ReconcileConfig,
    LakebridgeConfiguration,
    ReconcileMetadataConfig,
    TranspileConfig,
)
from databricks.labs.lakebridge.contexts.application import ApplicationContext
from databricks.labs.lakebridge.deployment.configurator import ResourceConfigurator
from databricks.labs.lakebridge.deployment.installation import WorkspaceInstallation
from databricks.labs.lakebridge.reconcile.constants import ReconReportType, ReconSourceType
from databricks.labs.lakebridge.transpiler.installers import (
    BladebridgeInstaller,
    MorpheusInstaller,
    TranspilerInstaller,
)
from databricks.labs.lakebridge.transpiler.repository import TranspilerRepository

logger = logging.getLogger(__name__)

TRANSPILER_WAREHOUSE_PREFIX = "Lakebridge Transpiler Validation"


class WorkspaceInstaller:
    # TODO: Temporary suppression, is_interactive is pending removal.
    def __init__(  # pylint: disable=too-many-arguments
        self,
        ws: WorkspaceClient,
        prompts: Prompts,
        installation: Installation,
        install_state: InstallState,
        product_info: ProductInfo,
        resource_configurator: ResourceConfigurator,
        workspace_installation: WorkspaceInstallation,
        environ: dict[str, str] | None = None,
        *,
        is_interactive: bool = True,
        include_llm: bool = False,
        switch_use_serverless: bool = True,
        transpiler_repository: TranspilerRepository = TranspilerRepository.user_home(),
        transpiler_installers: Sequence[Callable[[TranspilerRepository], TranspilerInstaller]] = (
            BladebridgeInstaller,
            MorpheusInstaller,
        ),
    ):
        self._ws = ws
        self._prompts = prompts
        self._installation = installation
        self._install_state = install_state
        self._product_info = product_info
        self._resource_configurator = resource_configurator
        self._ws_installation = workspace_installation
        # TODO: Refactor the 'prompts' property in preference to using this flag, which should be redundant.
        self._is_interactive = is_interactive
        self._include_llm = include_llm
        self._switch_use_serverless = switch_use_serverless
        self._transpiler_repository = transpiler_repository
        self._transpiler_installer_factories = transpiler_installers

        if not environ:
            environ = dict(os.environ.items())

        if "DATABRICKS_RUNTIME_VERSION" in environ:
            msg = "WorkspaceInstaller is not supposed to be executed in Databricks Runtime"
            raise SystemExit(msg)

    @property
    def _transpiler_installers(self) -> Set[TranspilerInstaller]:
        return frozenset(factory(self._transpiler_repository) for factory in self._transpiler_installer_factories)

    def run(
        self,
        module: str,
        config: LakebridgeConfiguration | None = None,
        artifact: str | None = None,
    ) -> LakebridgeConfiguration:
        logger.debug(f"Initializing workspace installation for module: {module} (config: {config})")
        if module == "transpile" and artifact:
            self._install_artifact(artifact)
        elif module in {"transpile", "all"}:
            for transpiler_installer in self._transpiler_installers:
                transpiler_installer.install()
        if not config:
            config = self.configure(module)
        if self._is_testing():
            return config
        self._ws_installation.install(config)
        logger.info("Installation completed successfully! Please refer to the documentation for the next steps.")
        return config

    def upgrade_installed_transpilers(self) -> bool:
        """Detect and upgrade, if possible and necessary, installed transpilers."""
        installed_transpilers = self._transpiler_repository.all_transpiler_names()
        if installed_transpilers:
            logger.info(f"Detected installed transpilers: {sorted(installed_transpilers)}")
        upgraded = False
        for transpiler_installer in self._transpiler_installers:
            name = transpiler_installer.name
            if name in installed_transpilers:
                logger.info(f"Checking for {name} upgrades...")
                upgraded |= transpiler_installer.install()
        # If we upgraded anything, the configuration process needs to run again.
        if upgraded:
            config = self.configure("transpile")
            if not self._is_testing():
                self._ws_installation.install(config)
        return upgraded

    def _install_artifact(self, artifact: str) -> None:
        path = Path(artifact)
        if not path.exists():
            logger.error(f"Could not locate artifact {artifact}")
            return
        for transpiler_installer in self._transpiler_installers:
            if transpiler_installer.can_install(path):
                transpiler_installer.install(path)
                break
        else:
            logger.fatal(f"Cannot install unsupported artifact: {artifact}")

    def configure(self, module: str) -> LakebridgeConfiguration:
        match module:
            case "transpile":
                logger.info("Configuring lakebridge `transpile`.")
                return LakebridgeConfiguration(
                    self._configure_transpile(),
                    reconcile=None,
                    include_switch=self._include_llm,
                    switch_use_serverless=self._switch_use_serverless,
                )
            case "reconcile":
                logger.info("Configuring lakebridge `reconcile`.")
                return LakebridgeConfiguration(None, self._configure_reconcile())
            case "all":
                logger.info("Configuring lakebridge `transpile` and `reconcile`.")
                return LakebridgeConfiguration(
                    self._configure_transpile(),
                    self._configure_reconcile(),
                    include_switch=self._include_llm,
                    switch_use_serverless=self._switch_use_serverless,
                )
            case _:
                raise ValueError(f"Invalid input: {module}")

    def _is_testing(self):
        return self._product_info.product_name() != "lakebridge"

    def _configure_transpile(self) -> TranspileConfig | None:
        try:
            config = self._installation.load(TranspileConfig)
            logger.info("Lakebridge `transpile` is already installed on this workspace.")
            if not self._is_interactive:
                logger.debug("Installation is not interactive, keeping existing configuration.")
                return config
            if not self._prompts.confirm("Do you want to override the existing installation?"):
                return config
        except NotFound:
            logger.info("Couldn't find existing `transpile` installation")
        except (PermissionDenied, SerdeError, ValueError, AttributeError):
            install_dir = self._installation.install_folder()
            logger.warning(
                f"Existing `transpile` installation at {install_dir} is corrupted. Continuing new installation..."
            )

        if not self._is_interactive:
            logger.warning("Installation is not interactive, skipping configuration of transpilers.")
            return None

        config = self._configure_new_transpile_installation()
        logger.info("Finished configuring lakebridge `transpile`.")
        return config

    def _configure_new_transpile_installation(self) -> TranspileConfig:
        default_config = self._prompt_for_new_transpile_installation()
        runtime_config = None
        catalog_name = "remorph"
        schema_name = "transpiler"
        if not default_config.skip_validation:
            catalog_name = self._configure_catalog()
            schema_name = self._configure_schema(catalog_name, schema_name)
            self._has_necessary_access(catalog_name, schema_name)
            warehouse_id = self._resource_configurator.prompt_for_warehouse_setup(TRANSPILER_WAREHOUSE_PREFIX)
            runtime_config = {"warehouse_id": warehouse_id}

        config = dataclasses.replace(
            default_config,
            catalog_name=catalog_name,
            schema_name=schema_name,
            sdk_config=runtime_config,
        )
        self._save_config(config)
        return config

    def _all_installed_dialects(self) -> list[str]:
        return sorted(self._transpiler_repository.all_dialects())

    def _transpilers_with_dialect(self, dialect: str) -> list[str]:
        return sorted(self._transpiler_repository.transpilers_with_dialect(dialect))

    def _transpiler_config_path(self, transpiler: str) -> Path:
        return self._transpiler_repository.transpiler_config_path(transpiler)

    # Sentinel value for special "set it later" option in prompts.
    _install_later = "Set it later"

    def _prompt_for_new_transpile_installation(self) -> TranspileConfig:
        # TODO tidy this up, logger might not display the below in console...
        logger.info("Please answer a few questions to configure lakebridge `transpile`")
        all_dialects = [self._install_later, *self._all_installed_dialects()]
        source_dialect: str | None = self._prompts.choice("Select the source dialect:", all_dialects, sort=False)
        if source_dialect == self._install_later:
            source_dialect = None
        transpiler_config_path: Path | None = None
        transpiler_options: dict[str, JsonValue] | None = None
        if source_dialect:
            transpilers = self._transpilers_with_dialect(source_dialect)
            if (found_config := self._get_transpiler_config(transpilers)) is not None:
                transpiler_name, transpiler_config_path = found_config
                transpiler_options = self._prompt_for_transpiler_options(transpiler_name, source_dialect)
        input_source: str | None = self._prompts.question(
            "Enter input SQL path (directory/file)", default=self._install_later
        )
        if input_source == self._install_later:
            input_source = None
        output_folder = self._prompts.question("Enter output directory", default="transpiled")
        # When defaults are passed along we need to use absolute paths to avoid issues with relative paths.
        if output_folder == "transpiled":
            output_folder = str(Path.cwd() / "transpiled")
        error_file_path = self._prompts.question("Enter error file path", default="errors.log")
        if error_file_path == "errors.log":
            error_file_path = str(Path.cwd() / "errors.log")

        run_validation = self._prompts.confirm(
            "Would you like to validate the syntax and semantics of the transpiled queries?"
        )

        return TranspileConfig(
            transpiler_config_path=str(transpiler_config_path) if transpiler_config_path is not None else None,
            transpiler_options=transpiler_options,
            source_dialect=source_dialect,
            skip_validation=(not run_validation),
            input_source=input_source,
            output_folder=output_folder,
            error_file_path=error_file_path,
        )

    def _get_transpiler_config(self, transpiler_names: list[str]) -> tuple[str, Path] | None:
        match transpiler_names:
            case [only_name]:
                transpiler_name = only_name
                logger.info(f"Lakebridge will use the {transpiler_name} transpiler")
            case _:
                choices = [self._install_later, *transpiler_names]
                transpiler_name = self._prompts.choice("Select the transpiler:", choices, sort=False)
                if transpiler_name == self._install_later:
                    return None
        transpiler_config_path = self._transpiler_config_path(transpiler_name)
        return transpiler_name, transpiler_config_path

    def _prompt_for_transpiler_options(self, transpiler_name: str, source_dialect: str) -> dict[str, Any] | None:
        config_options = self._transpiler_repository.transpiler_config_options(transpiler_name, source_dialect)
        if len(config_options) == 0:
            return None
        # Semantics here are different from the other properties of a TranspileConfig. Specifically:
        #  - Entries are present for all options.
        #  - If the value is None, it means the user chose to not provide a value. (This differs to other
        #    attributes, where None means the user chose to provide a value later.)
        #  - There is no way to express 'provide a value later'.
        return {option.flag: option.prompt_for_value(self._prompts) for option in config_options}

    def _configure_catalog(self) -> str:
        return self._resource_configurator.prompt_for_catalog_setup()

    def _configure_schema(
        self,
        catalog: str,
        default_schema_name: str,
    ) -> str:
        return self._resource_configurator.prompt_for_schema_setup(
            catalog,
            default_schema_name,
        )

    def _configure_reconcile(self) -> ReconcileConfig:
        try:
            self._installation.load(ReconcileConfig)
            logger.info("Lakebridge `reconcile` is already installed on this workspace.")
            if not self._prompts.confirm("Do you want to override the existing installation?"):
                # TODO: Exit gracefully, without raising SystemExit
                raise SystemExit(
                    "Lakebridge `reconcile` is already installed and no override has been requested. Exiting..."
                )
        except NotFound:
            logger.info("Couldn't find existing `reconcile` installation")
        except (PermissionDenied, SerdeError, ValueError, AttributeError):
            install_dir = self._installation.install_folder()
            logger.warning(
                f"Existing `reconcile` installation at {install_dir} is corrupted. Continuing new installation..."
            )

        config = self._configure_new_reconcile_installation()
        logger.info("Finished configuring lakebridge `reconcile`.")
        return config

    def _configure_new_reconcile_installation(self) -> ReconcileConfig:
        default_config = self._prompt_for_new_reconcile_installation()
        self._save_config(default_config)
        return default_config

    def _prompt_for_new_reconcile_installation(self) -> ReconcileConfig:
        logger.info("Please answer a few questions to configure lakebridge `reconcile`")
        data_source = self._prompts.choice(
            "Select the Data Source:", [source_type.value for source_type in ReconSourceType]
        )
        report_type = self._prompts.choice(
            "Select the report type:", [report_type.value for report_type in ReconReportType]
        )
        scope_name = self._prompts.question(
            f"Enter Secret scope name to store `{data_source.capitalize()}` connection details / secrets",
            default=f"remorph_{data_source}",
        )

        db_config = self._prompt_for_reconcile_database_config(data_source)
        metadata_config = self._prompt_for_reconcile_metadata_config()

        return ReconcileConfig(
            data_source=data_source,
            report_type=report_type,
            secret_scope=scope_name,
            database_config=db_config,
            metadata_config=metadata_config,
        )

    def _prompt_for_reconcile_database_config(self, source) -> DatabaseConfig:
        source_catalog = None
        if source == ReconSourceType.SNOWFLAKE.value:
            source_catalog = self._prompts.question(f"Enter source catalog name for `{source.capitalize()}`")
        if source == ReconSourceType.DATABRICKS.value:
            source_catalog = self._prompts.question(
                f"Enter source catalog name for `{source.capitalize()}`", default="hive_metastore"
            )

        schema_prompt = f"Enter source schema name for `{source.capitalize()}`"
        if source in {ReconSourceType.ORACLE.value}:
            schema_prompt = f"Enter source database name for `{source.capitalize()}`"

        source_schema = self._prompts.question(schema_prompt)
        target_catalog = self._prompts.question("Enter target catalog name for Databricks")
        target_schema = self._prompts.question("Enter target schema name for Databricks")

        return DatabaseConfig(
            source_schema=source_schema,
            target_catalog=target_catalog,
            target_schema=target_schema,
            source_catalog=source_catalog,
        )

    def _prompt_for_reconcile_metadata_config(self) -> ReconcileMetadataConfig:
        logger.info("Configuring reconcile metadata.")
        catalog = self._configure_catalog()
        schema = self._configure_schema(
            catalog,
            "reconcile",
        )
        volume = self._configure_volume(catalog, schema, "reconcile_volume")
        self._has_necessary_access(catalog, schema, volume)
        return ReconcileMetadataConfig(catalog=catalog, schema=schema, volume=volume)

    def _configure_volume(
        self,
        catalog: str,
        schema: str,
        default_volume_name: str,
    ) -> str:
        return self._resource_configurator.prompt_for_volume_setup(
            catalog,
            schema,
            default_volume_name,
        )

    def _save_config(self, config: TranspileConfig | ReconcileConfig):
        logger.info(f"Saving configuration file {config.__file__}")
        self._installation.save(config)
        ws_file_url = self._installation.workspace_link(config.__file__)
        if self._prompts.confirm(f"Open config file {ws_file_url} in the browser?"):
            webbrowser.open(ws_file_url)

    def _has_necessary_access(self, catalog_name: str, schema_name: str, volume_name: str | None = None):
        self._resource_configurator.has_necessary_access(catalog_name, schema_name, volume_name)


def installer(
    ws: WorkspaceClient,
    transpiler_repository: TranspilerRepository,
    *,
    is_interactive: bool,
    include_llm: bool = False,
    switch_use_serverless: bool = True,
) -> WorkspaceInstaller:
    app_context = ApplicationContext(_verify_workspace_client(ws))
    return WorkspaceInstaller(
        app_context.workspace_client,
        app_context.prompts,
        app_context.installation,
        app_context.install_state,
        app_context.product_info,
        app_context.resource_configurator,
        app_context.workspace_installation,
        transpiler_repository=transpiler_repository,
        is_interactive=is_interactive,
        include_llm=include_llm,
        switch_use_serverless=switch_use_serverless,
    )


def _verify_workspace_client(ws: WorkspaceClient) -> WorkspaceClient:
    """Verifies the workspace client configuration, ensuring it has the correct product info."""

    # Using reflection to set right value for _product_info for telemetry
    product_info = getattr(ws.config, '_product_info')
    if product_info[0] != "lakebridge":
        setattr(ws.config, '_product_info', ('lakebridge', __version__))

    return ws


if __name__ == "__main__":
    with_user_agent_extra("cmd", "install")
    initialize_logging()

    # Warning: ensures logger for this file is not __main__.
    logger = get_logger(__file__)

    app_installer = installer(
        ws=lakebridge.create_workspace_client(),
        transpiler_repository=TranspilerRepository.user_home(),
        is_interactive=sys.stdin.isatty(),
    )
    if not app_installer.upgrade_installed_transpilers():
        logger.debug("No existing Lakebridge transpilers detected; assuming fresh installation.")

    logger.info("Successfully Setup Lakebridge Components Locally")
    logger.info("For more information, please visit https://databrickslabs.github.io/lakebridge/")



from pathlib import Path
import logging
from typing import Any, Protocol

import yaml

from databricks.labs.lakebridge.connections.env_getter import EnvGetter

logger = logging.getLogger(__name__)


class SecretProvider(Protocol):
    def get_secret(self, key: str) -> str:
        pass


class LocalSecretProvider(SecretProvider):
    def get_secret(self, key: str) -> str:
        return key


class EnvSecretProvider(SecretProvider):
    def __init__(self, env_getter: EnvGetter):
        self._env_getter = env_getter

    def get_secret(self, key: str) -> str:
        try:
            return self._env_getter.get(str(key))
        except KeyError:
            logger.debug(f"Environment variable {key} not found. Falling back to actual value")
            return key


class DatabricksSecretProvider:
    def get_secret(self, key: str) -> str:
        raise NotImplementedError("Databricks secret vault not implemented")


class CredentialManager:
    def __init__(self, credentials: dict, secret_providers: dict[str, SecretProvider]):
        self._credentials = credentials
        self._default_vault = self._credentials.get('secret_vault_type', 'local').lower()
        self._provider = secret_providers.get(self._default_vault)
        if not self._provider:
            raise ValueError(f"Unsupported secret vault type: {self._default_vault}")

    def get_credentials(self, source: str) -> dict[str, Any]:
        if source not in self._credentials:
            raise KeyError(f"Source system: {source} credentials not found")

        value = self._credentials[source]
        if not isinstance(value, dict):
            raise KeyError(f"Invalid credential format for source: {source}")

        # Safe to cast: we verified value is a dict, so _resolve_credentials returns a dict
        return self._resolve_credentials(value)

    def _resolve_credentials(self, value: dict[str, Any]) -> dict[str, Any]:
        """Recursively resolve credentials, handling nested dictionaries and secret values.

        rules:
        - dict: Recursively process each key-value pair
        - list: Recursively process each item
        - str: Apply secret provider (resolve from env vars or return as-is)
        - Other types (int, bool, None, float): Return unchanged

            Processed value with the same structure but string values resolved
        """
        if isinstance(value, dict):
            return {k: self._resolve_credentials(v) for k, v in value.items()}
        if isinstance(value, list):
            return [self._resolve_credentials(item) for item in value]
        if isinstance(value, str):
            return self._get_secret_value(value)
        # For int, bool, None, float, etc., return as-is
        return value

    def _get_secret_value(self, key: str) -> str:
        """Apply the configured secret provider to resolve a string value."""
        assert self._provider is not None
        return self._provider.get_secret(key)


def _get_home() -> Path:
    return Path.home()


def cred_file(product_name) -> Path:
    return _get_home() / ".databricks" / "labs" / product_name / ".credentials.yml"


def _load_credentials(path: Path) -> dict:
    try:
        with open(path, encoding="utf-8") as f:
            return yaml.safe_load(f)
    except FileNotFoundError as e:
        raise FileNotFoundError(f"Credentials file not found at {path}") from e


def create_credential_manager(
    product_name: str,
    env_getter: EnvGetter,
    creds_path: Path | None = None,
) -> CredentialManager:
    if creds_path is None:
        creds_path = cred_file(product_name)
    creds = _load_credentials(creds_path)

    secret_providers = {
        'local': LocalSecretProvider(),
        'env': EnvSecretProvider(env_getter),
        'databricks': DatabricksSecretProvider(),
    }

    return CredentialManager(creds, secret_providers)



import re
import logging
from collections.abc import Mapping
from datetime import datetime

from pyspark.errors import PySparkException
from pyspark.sql import DataFrame, DataFrameReader, SparkSession
from pyspark.sql.functions import col
from sqlglot import Dialect

from databricks.labs.lakebridge.reconcile.connectors.data_source import DataSource
from databricks.labs.lakebridge.reconcile.connectors.jdbc_reader import JDBCReaderMixin
from databricks.labs.lakebridge.reconcile.connectors.models import NormalizedIdentifier
from databricks.labs.lakebridge.reconcile.connectors.secrets import SecretsMixin
from databricks.labs.lakebridge.reconcile.connectors.dialect_utils import DialectUtils
from databricks.labs.lakebridge.reconcile.recon_config import JdbcReaderOptions, Schema, OptionalPrimitiveType
from databricks.sdk import WorkspaceClient

logger = logging.getLogger(__name__)


class OracleDataSource(DataSource, SecretsMixin, JDBCReaderMixin):
    _DRIVER = "oracle"
    _IDENTIFIER_DELIMITER = "\""
    _SCHEMA_QUERY = """select column_name, case when (data_precision is not null
                                              and data_scale <> 0)
                                              then data_type || '(' || data_precision || ',' || data_scale || ')'
                                              when (data_precision is not null and data_scale = 0)
                                              then data_type || '(' || data_precision || ')'
                                              when data_precision is null and (lower(data_type) in ('date') or
                                              lower(data_type) like 'timestamp%') then  data_type
                                              when CHAR_LENGTH = 0 then data_type
                                              else data_type || '(' || CHAR_LENGTH || ')'
                                              end data_type
                                              FROM ALL_TAB_COLUMNS
                            WHERE lower(TABLE_NAME) = '{table}' and lower(owner) = '{owner}'"""

    def __init__(
        self,
        engine: Dialect,
        spark: SparkSession,
        ws: WorkspaceClient,
        secret_scope: str,
    ):
        self._engine = engine
        self._spark = spark
        self._ws = ws
        self._secret_scope = secret_scope

    @property
    def get_jdbc_url(self) -> str:
        return (
            f"jdbc:{OracleDataSource._DRIVER}:thin:@//{self._get_secret('host')}"
            f":{self._get_secret('port')}/{self._get_secret('database')}"
        )

    def read_data(
        self,
        catalog: str | None,
        schema: str,
        table: str,
        query: str,
        options: JdbcReaderOptions | None,
    ) -> DataFrame:
        table_query = query.replace(":tbl", f"{schema}.{table}")
        try:
            if options is None:
                return self.reader(table_query, self._get_timestamp_options()).load()
            reader_options = self._get_jdbc_reader_options(options) | self._get_timestamp_options()
            df = self.reader(table_query, reader_options).load()
            logger.warning(f"Fetching data using query: \n`{table_query}`")

            # Convert all column names to lower case
            df = df.select([col(c).alias(c.lower()) for c in df.columns])
            return df
        except (RuntimeError, PySparkException) as e:
            return self.log_and_throw_exception(e, "data", table_query)

    def get_schema(
        self,
        catalog: str | None,
        schema: str,
        table: str,
        normalize: bool = True,
    ) -> list[Schema]:
        schema_query = re.sub(
            r'\s+',
            ' ',
            OracleDataSource._SCHEMA_QUERY.format(table=table.lower(), owner=schema.lower()),
        )
        try:
            logger.debug(f"Fetching schema using query: \n`{schema_query}`")
            logger.info(f"Fetching Schema: Started at: {datetime.now()}")
            df = self.reader(schema_query).load()
            schema_metadata = df.select([col(c).alias(c.lower()) for c in df.columns]).collect()
            logger.info(f"Schema fetched successfully. Completed at: {datetime.now()}")
            logger.debug(f"schema_metadata: {schema_metadata}")
            return [self._map_meta_column(field, normalize) for field in schema_metadata]
        except (RuntimeError, PySparkException) as e:
            return self.log_and_throw_exception(e, "schema", schema_query)

    @staticmethod
    def _get_timestamp_options() -> dict[str, str]:
        return {
            "oracle.jdbc.mapDateToTimestamp": "false",
            "sessionInitStatement": "BEGIN dbms_session.set_nls('nls_date_format', "
            "'''YYYY-MM-DD''');dbms_session.set_nls('nls_timestamp_format', '''YYYY-MM-DD "
            "HH24:MI:SS''');END;",
        }

    def reader(self, query: str, options: Mapping[str, OptionalPrimitiveType] | None = None) -> DataFrameReader:
        if options is None:
            options = {}
        user = self._get_secret('user')
        password = self._get_secret('password')
        logger.debug(f"Using user: {user} to connect to Oracle")
        return self._get_jdbc_reader(
            query, self.get_jdbc_url, OracleDataSource._DRIVER, {**options, "user": user, "password": password}
        )

    def normalize_identifier(self, identifier: str) -> NormalizedIdentifier:
        normalized = DialectUtils.normalize_identifier(
            identifier,
            source_start_delimiter=OracleDataSource._IDENTIFIER_DELIMITER,
            source_end_delimiter=OracleDataSource._IDENTIFIER_DELIMITER,
        )

        # TODO: In Oracle, quoted identifiers are case-sensitive,
        # it is disabled for now till we have a proper strategy to handle it.
        normalized.source_normalized = DialectUtils.unnormalize_identifier(normalized.ansi_normalized)

        return normalized



import os


class EnvGetter:
    """Standardised inorder to support testing Capabilities, check debug_envgetter.py"""

    def __init__(self):
        self.env = dict(os.environ)

    def get(self, key: str) -> str:
        if key in self.env:
            return self.env[key]
        raise KeyError(f"not in env: {key}")



import io
import os
import json

import logging
from pathlib import Path

from databricks.sdk.errors import PermissionDenied, NotFound, InternalError
from databricks.sdk.errors.platform import ResourceAlreadyExists, DatabricksError
from databricks.sdk.service.dashboards import Dashboard
from databricks.sdk import WorkspaceClient

from databricks.labs.blueprint.installation import Installation
from databricks.labs.blueprint.installer import InstallState
from databricks.labs.blueprint.wheels import find_project_root

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class DashboardTemplateLoader:
    """
    Class for loading the JSON representation of a Databricks dashboard
    according to the source system.
    """

    def __init__(self, templates_dir: Path | None):
        self.templates_dir = templates_dir

    def load(self, source_system: str) -> dict:
        """
        Loads a profiler summary dashboard.
        :param source_system: - the name of the source data warehouse
        """
        if self.templates_dir is None:
            raise ValueError("Dashboard template path cannot be empty.")

        filename = f"{source_system.lower()}_dashboard.lvdash.json"
        filepath = os.path.join(self.templates_dir, filename)
        if not os.path.exists(filepath):
            raise FileNotFoundError(f"Could not find dashboard template matching {source_system}.")
        with open(filepath, "r", encoding="utf-8") as f:
            return json.load(f)


class DashboardManager:
    """
    Class for managing the lifecycle of a profiler dashboard summary, a.k.a. "local dashboards"
    """

    _DASHBOARD_NAME = "Lakebridge Profiler Assessment"

    def __init__(
        self, ws: WorkspaceClient, installation: Installation, install_state: InstallState, is_debug: bool = False
    ):
        self._ws = ws
        self._installation = installation
        self._install_state = install_state
        self._is_debug = is_debug

    @staticmethod
    def _replace_catalog_schema(
        serialized_dashboard: str,
        new_catalog: str,
        new_schema: str,
        old_catalog: str = "`PROFILER_CATALOG`",
        old_schema: str = "`PROFILER_SCHEMA`",
    ):
        """Given a serialized JSON dashboard, replaces all catalog and schema references with the
        provided catalog and schema names."""
        updated_dashboard = serialized_dashboard.replace(old_catalog, f"`{new_catalog}`")
        return updated_dashboard.replace(old_schema, f"`{new_schema}`")

    def _create_or_replace_dashboard(
        self, folder: Path, ws_parent_path: str, dest_catalog: str, dest_schema: str
    ) -> Dashboard:
        """
        Creates or updates a profiler summary dashboard in the current user’s Databricks workspace home.
        Existing dashboards are automatically replaced with the latest dashboard template.
        """

        # Load the dashboard template
        logging.info(f"Loading dashboard template from folder: {folder}")
        dash_reference = f"{folder.stem}".lower()
        dashboard_loader = DashboardTemplateLoader(folder)
        dashboard_json = dashboard_loader.load(source_system="synapse")
        dashboard_str = json.dumps(dashboard_json)

        # Replace catalog and schema placeholders
        updated_dashboard_str = self._replace_catalog_schema(
            dashboard_str, new_catalog=dest_catalog, new_schema=dest_schema
        )
        dashboard = Dashboard(
            display_name=self._DASHBOARD_NAME,
            parent_path=ws_parent_path,
            warehouse_id=self._ws.config.warehouse_id,
            serialized_dashboard=updated_dashboard_str,
        )

        # Create dashboard or replace if previously deployed
        try:
            dashboard = self._ws.lakeview.create(dashboard=dashboard)
        except ResourceAlreadyExists:
            logging.info("Dashboard already exists! Removing dashboard from workspace location.")
            dashboard_ws_path = str(Path(ws_parent_path) / f"{self._DASHBOARD_NAME}.lvdash.json")
            self._ws.workspace.delete(dashboard_ws_path)
            dashboard = self._ws.lakeview.create(dashboard=dashboard)
        except DatabricksError as e:
            logging.error(f"Could not create profiler summary dashboard: {e}")

        assert dashboard.dashboard_id is not None
        logging.info(f"Created dashboard '{dashboard.dashboard_id}' in workspace location {ws_parent_path}.")
        self._install_state.dashboards[dash_reference] = dashboard.dashboard_id
        return dashboard

    def create_profiler_summary_dashboard(
        self,
        source_tech: str,
        catalog_name: str = "lakebridge_profiler",
        schema_name: str = "profiler_runs",
    ) -> None:
        """Deploys a profiler summary dashboard to the current Databricks user’s workspace home."""

        logger.info("Deploying profiler summary dashboard.")

        # Load the dashboard template for the source system
        template_folder = (
            find_project_root(__file__)
            / f"src/databricks/labs/lakebridge/resources/assessments/dashboards/{source_tech}"
        )
        logger.info(f"Deploying profiler dashboard from template folder: {template_folder}")
        ws_parent_path = f"{self._installation.install_folder()}/dashboards"
        try:
            self._ws.workspace.mkdirs(ws_parent_path)
        except ResourceAlreadyExists:
            logger.info(f"Workspace parent path already exists for dashboards: {ws_parent_path}")
        self._create_or_replace_dashboard(
            folder=template_folder, ws_parent_path=ws_parent_path, dest_catalog=catalog_name, dest_schema=schema_name
        )

    def upload_duckdb_to_uc_volume(self, local_file_path, volume_path):
        """
        Upload a DuckDB file to Unity Catalog Volume
        Args:
            local_file_path (str): Local path to the DuckDB file
            volume_path (str): Target path in UC Volume (e.g., '/Volumes/catalog/schema/volume/myfile.duckdb')
        Returns:
            bool: True if successful, False otherwise
        """

        # Validate inputs
        if not os.path.exists(local_file_path):
            logger.error(f"Local file not found: {local_file_path}")
            return False

        if not volume_path.startswith('/Volumes/'):
            logger.error("Volume path must start with '/Volumes/'")
            return False

        try:
            with open(local_file_path, 'rb') as f:
                file_bytes = f.read()
                binary_data = io.BytesIO(file_bytes)
                self._ws.files.upload(volume_path, binary_data, overwrite=True)
            logger.info(f"Successfully uploaded {local_file_path} to {volume_path}")
            return True
        except FileNotFoundError as e:
            logger.error(f"Profiler extract file was not found: \n{e}")
            return False
        except PermissionDenied as e:
            logger.error(f"Insufficient privileges detected while accessing Volume path: \n{e}")
            return False
        except NotFound as e:
            logger.error(f"Invalid Volume path provided: \n{e}")
            return False
        except InternalError as e:
            logger.error(f"Internal Databricks error while uploading extract file: \n{e}")
            return False



import json
import sys

import duckdb
from databricks.labs.blueprint.entrypoint import get_logger

from databricks.labs.lakebridge import initialize_logging
from databricks.labs.lakebridge.assessments import PRODUCT_NAME
from databricks.labs.lakebridge.connections.credential_manager import create_credential_manager
from databricks.labs.lakebridge.connections.env_getter import EnvGetter
from databricks.labs.lakebridge.connections.synapse_connection_helpers import create_synapse_connection
from databricks.labs.lakebridge.resources.assessments.synapse.common.duckdb_helpers import (
    save_resultset_to_db,
    get_max_column_value_duckdb,
)
from databricks.labs.lakebridge.resources.assessments.synapse.common.functions import arguments_loader
from databricks.labs.lakebridge.resources.assessments.synapse.common.queries import SynapseQueries


logger = get_logger(__file__)


def get_serverless_database_groups(
    db_path,
    inclusion_list=None,
    exclusion_list=None,
    table_name="serverless_databases",
):
    """
    Reads serverless databases from DuckDB, groups by collation_name,
    and applies inclusion/exclusion filters.

    Returns:
        (serverless_database_groups, serverless_db_groups_in_scope)
    """
    with duckdb.connect(db_path) as conn:
        rows = conn.execute(f"SELECT name, collation_name FROM {table_name}").fetchall()

    serverless_database_groups = {}
    for name, collation_name in rows:
        serverless_database_groups.setdefault(collation_name, []).append(name)

    inclusion_list = inclusion_list or []
    exclusion_list = exclusion_list or []

    serverless_db_groups_in_scope = {}
    for collation_name, dbs in serverless_database_groups.items():
        for database in dbs:
            if (not inclusion_list or database in inclusion_list) and database not in exclusion_list:
                serverless_db_groups_in_scope.setdefault(collation_name, []).append(database)

    return serverless_db_groups_in_scope


def execute():
    db_path, creds_file = arguments_loader(desc="Synapse Synapse Serverless SQL Pool Extract Script")

    cred_manager = create_credential_manager(PRODUCT_NAME, EnvGetter())
    synapse_workspace_settings = cred_manager.get_credentials("synapse")
    config = synapse_workspace_settings["workspace"]
    auth_type = synapse_workspace_settings["jdbc"].get("auth_type", "sql_authentication")
    synapse_profiler_settings = synapse_workspace_settings["profiler"]

    connection = None
    try:
        if not synapse_profiler_settings.get("exclude_serverless_sql_pool", False):
            # Databases
            database_query = SynapseQueries.list_databases()
            connection = create_synapse_connection(
                workspace_config=config,
                database='master',
                endpoint_key='serverless_sql_endpoint',
                auth_type=auth_type,
            )
            result = connection.fetch(database_query)
            save_resultset_to_db(result, "serverless_databases", db_path, mode="overwrite")

            serverless_db_groups_in_scope = get_serverless_database_groups(db_path)
            logger.info(f"serverless db in scope: {serverless_db_groups_in_scope}")

            for idx, collation_name in enumerate(serverless_db_groups_in_scope):
                mode = "overwrite" if idx == 0 else "append"
                databases = serverless_db_groups_in_scope[collation_name]

                for db_name in databases:
                    connection = create_synapse_connection(
                        workspace_config=config,
                        database=db_name,
                        endpoint_key='serverless_sql_endpoint',
                        auth_type=auth_type,
                    )
                    # tables
                    table_name = 'serverless_tables'
                    table_query = SynapseQueries.list_tables(db_name)
                    logger.info(f"Loading '{table_name}' for pool: %s", db_name)
                    result = connection.fetch(table_query)
                    save_resultset_to_db(result, table_name, db_path, mode=mode)

                    # columns
                    table_name = "serverless_columns"
                    column_query = SynapseQueries.list_columns(db_name)
                    logger.info(f"Loading '{table_name}' for pool: %s", db_name)
                    result = connection.fetch(column_query)
                    save_resultset_to_db(result, table_name, db_path, mode=mode)

                    # views
                    table_name = "serverless_views"
                    view_query = SynapseQueries.list_views(db_name)
                    logger.info(f"Loading '{table_name}' for pool: %s", db_name)
                    result = connection.fetch(view_query)
                    save_resultset_to_db(result, table_name, db_path, mode=mode)

                    # Routines
                    table_name = "serverless_routines"
                    view_query = SynapseQueries.list_serverless_routines(db_name, True)
                    logger.info(f"Loading '{table_name}' for pool: %s", db_name)
                    result = connection.fetch(view_query)
                    save_resultset_to_db(result, table_name, db_path, mode=mode)

                    mode = "append"

            # Reconnect to master for server-level DMVs
            connection = create_synapse_connection(
                workspace_config=config,
                database='master',
                endpoint_key='serverless_sql_endpoint',
                auth_type=auth_type,
            )

            pool_name = "serverless"
            # Data Processed - server-level DMV, must be queried from master database
            table_name = "serverless_data_processed"
            logger.info(f"Loading '{table_name}' for pool: %s", pool_name)
            data_processed_query = SynapseQueries.data_processed(pool_name)

            session_result = connection.fetch(data_processed_query)
            save_resultset_to_db(session_result, table_name, db_path, mode="append")

            # Activity Extract:
            table_name = "serverless_sessions"
            logger.info(f"Loading '{table_name}' for pool: %s", pool_name)
            prev_max_login_time = get_max_column_value_duckdb("login_time", table_name, db_path)
            session_query = SynapseQueries.list_serverless_sessions(pool_name, prev_max_login_time)

            session_result = connection.fetch(session_query)
            save_resultset_to_db(session_result, table_name, db_path, mode="append")

            table_name = "serverless_session_requests"
            logger.info(f"Loading '{table_name}' for pool: %s", pool_name)
            prev_max_end_time = get_max_column_value_duckdb("start_time", table_name, db_path)
            session_request_query = SynapseQueries.list_serverless_requests(pool_name, prev_max_end_time)

            session_request_result = connection.fetch(session_request_query)
            save_resultset_to_db(session_request_result, table_name, db_path, mode="append")

            table_name = "serverless_query_stats"
            logger.info(f"Loading '{table_name}' for pool: %s", pool_name)
            max_last_execution_time = get_max_column_value_duckdb("last_execution_time", table_name, db_path)
            query_stats = SynapseQueries.list_query_stats(max_last_execution_time)

            session_result = connection.fetch(query_stats)
            save_resultset_to_db(session_result, table_name, db_path, mode="append")

            table_name = "serverless_requests_history"
            logger.info(f"Loading '{table_name}' for pool: %s", pool_name)
            max_end_time = get_max_column_value_duckdb("end_time", table_name, db_path)
            query_history = SynapseQueries.query_requests_history(max_end_time)

            session_request_result = connection.fetch(query_history)
            save_resultset_to_db(session_request_result, table_name, db_path, mode="append")

        else:
            logger.info("'exclude_serverless_sql_pool' configuration is set to True.Skipping Serverless Pool extracts.")

        print(json.dumps({"status": "success", "message": "Data loaded successfully"}))

    except Exception as e:
        logger.error(f"Failed to extract info for Synapse Serverless SQL Pool: {str(e)}")
        print(json.dumps({"status": "error", "message": str(e)}), file=sys.stderr)
        sys.exit(1)


if __name__ == '__main__':
    initialize_logging()
    execute()



from __future__ import annotations
import abc
from collections.abc import Sequence
from pathlib import Path

from databricks.labs.lakebridge.config import TranspileResult, TranspileConfig


class TranspileEngine(abc.ABC):
    @abc.abstractmethod
    async def initialize(self, config: TranspileConfig) -> None: ...

    @abc.abstractmethod
    async def shutdown(self) -> None: ...

    @abc.abstractmethod
    async def transpile(
        self, source_dialect: str, target_dialect: str, source_code: str, file_path: Path
    ) -> TranspileResult: ...

    @property
    @abc.abstractmethod
    def transpiler_name(self) -> str: ...

    @property
    @abc.abstractmethod
    def supported_dialects(self) -> Sequence[str]: ...

    @abc.abstractmethod
    def is_supported_file(self, file: Path) -> bool: ...



from __future__ import annotations

from collections.abc import Generator, Iterable, Mapping, Sequence, Set
from dataclasses import dataclass
from json import loads
from typing import Any
from pathlib import Path
import logging

from databricks.labs.lakebridge.config import LSPConfigOptionV1
from databricks.labs.lakebridge.transpiler.lsp.lsp_engine import LSPConfig

logger = logging.getLogger(__name__)


@dataclass(frozen=True, kw_only=True)
class TranspilerInfo:
    transpiler_name: str
    version: str | None
    configuration_path: Path
    dialects: Mapping[str, Sequence[LSPConfigOptionV1]]


class TranspilerRepository:
    """
    Repository for managing the installed transpilers in the user's home directory.

    The default repository for a user is always located under ~/.databricks/labs, and can be obtained
    via the `TranspilerRepository.user_home()` method.
    """

    #
    # Transpilers currently have different identifiers, for historical reasons:
    #
    #  - transpiler_id: the unique identifier of the transpiler according to this project, assigned within the `installer`
    #        module and used as the name of the directory into which the transpiler is installed within a repository.
    #  - transpiler_name: the name of the transpiler according to its own metadata, found in the configuration file
    #       bundled within each transpiler as distributed.
    #
    # Note: multiple installed transpilers might have the same transpiler name, but a transpiler id is unique to a single
    # installed transpiler.
    #
    #  Known identifiers at the moment:
    #
    #   - Morpheus:     transpiler_id = databricks-morph-plugin,  transpiler_name = Morpheus
    #   - BladeBridge:  transpiler_id = bladebridge,              transpiler_name = Bladebridge
    #

    @staticmethod
    def default_labs_path() -> Path:
        """Return the default path where labs applications are installed."""
        return Path.home() / ".databricks" / "labs"

    _default_repository: TranspilerRepository | None = None

    @classmethod
    def user_home(cls) -> TranspilerRepository:
        """The default repository for transpilers in the current user's home directory."""
        repository = cls._default_repository
        if repository is None:
            cls._default_repository = repository = cls(cls.default_labs_path())
        return repository

    def __init__(self, labs_path: Path) -> None:
        """Initialize the repository, based in the given location.

        This should only be used directly by tests; for the default repository, use `TranspilerRepository.user_home()`.

        Args:
            labs_path: The path where the labs applications are installed.
        """
        if self._default_repository == self and labs_path == self.default_labs_path():
            raise ValueError("Use TranspilerRepository.user_home() to get the default repository.")
        self._labs_path = labs_path

    def __repr__(self) -> str:
        return f"TranspilerRepository(labs_path={self._labs_path!r})"

    def transpilers_path(self) -> Path:
        return self._labs_path / "remorph-transpilers"

    @classmethod
    def _parse_version_file(cls, transpiler_path: Path) -> str | None:
        """
        Obtain the version of an installed transpiler.

        Args:
          transpiler_path: The path of the transpiler whose version is sought.
        Returns:
          The version of the transpiler if it is installed, or None otherwise.
        """
        current_version_path = transpiler_path / "state" / "version.json"
        try:
            text = current_version_path.read_text("utf-8")
        except FileNotFoundError:
            return None
        data: dict[str, Any] = loads(text)
        version: str | None = data.get("version", None)
        if not version or not version.startswith("v"):
            return None
        return version[1:]

    def get_installed_version(self, transpiler_id: str) -> str | None:
        # Warning: transpiler_id here (eg. 'morpheus') and transpiler_name elsewhere (eg. Morpheus) are not the same!
        transpiler_path = self.transpilers_path() / transpiler_id
        return self._parse_version_file(transpiler_path)

    def get_installed_version_given_config_path(self, transpiler_config_path: Path) -> str | None:
        transpiler_path = transpiler_config_path.parent.parent
        return self._parse_version_file(transpiler_path)

    def all_transpiler_configs(self) -> Mapping[str, LSPConfig]:
        """Obtain all installed transpile configurations.

        Returns:
          A mapping of configurations, keyed by their ids.
        """
        return {path.name: config for path, config in self._all_transpiler_configs()}

    def all_transpiler_names(self) -> Set[str]:
        """Query the set of transpiler names for all installed transpilers."""
        all_configs = self._all_transpiler_configs()
        return frozenset(config.name for _, config in all_configs)

    def _transpiler_locations(self) -> Generator[Path, None, None]:
        transpilers_path = self.transpilers_path()
        try:
            # Treat the first entry specially: failure here is different from failure once underway.
            iterator = transpilers_path.iterdir()
            yield next(iterator)
        except StopIteration:
            # Harmless: no transpilers installed.
            return
        except OSError as e:
            # Also generally non-fatal, on the first entry: normally means there's no installation.
            logger.debug("Unable to list installed transpilers", exc_info=e)
            return
        # After the first entry, continue yielding but any errors need to propagate: something is wrong.
        yield from iterator

    def installed_transpilers(self) -> Mapping[str, TranspilerInfo]:
        """Query the set of installed transpilers and their metadata."""
        return {
            path.name: TranspilerInfo(
                transpiler_name=config.name,
                version=self.get_installed_version(path.name),
                configuration_path=config.path,
                dialects={dialect: config.options_for_dialect(dialect) for dialect in config.remorph.dialects},
            )
            for path, config in self._all_transpiler_configs()
        }

    def all_dialects(self) -> Set[str]:
        """Query the set of dialects for all installed transpilers."""
        all_dialects: set[str] = set()
        for _, config in self._all_transpiler_configs():
            all_dialects = all_dialects.union(config.remorph.dialects)
        return all_dialects

    def transpilers_with_dialect(self, dialect: str) -> Set[str]:
        """
        Query the set of transpilers that can handle a given dialect.

        Args:
          dialect: The dialect to check for.
        Returns:
          The set of transpiler names for installed transpilers that can handle a given dialect.
        """
        configs = filter(lambda cfg: dialect in cfg.remorph.dialects, self.all_transpiler_configs().values())
        return frozenset(config.name for config in configs)

    def _find_transpile_config(self, transpiler_name: str) -> LSPConfig | None:
        try:
            return next(c for _, c in self._all_transpiler_configs() if c.name == transpiler_name)
        except StopIteration:
            return None

    def transpiler_config_path(self, transpiler_name: str) -> Path:
        """
        Obtain the path to a configuration file for an installed transpiler.

        Args:
          transpiler_name: The transpiler name of the transpiler whose path is sought.
        Returns:
          The path of the configuration file for the given installed transpiler. If multiple installed transpilers
          have the same transpiler name, either may be returned.
        Raises:
          ValueError: If there is no installed transpiler with the given transpiler name.
        """
        # Note: Because it's the transpiler name, have to hunt through the installed list rather get it directly.
        config = self._find_transpile_config(transpiler_name)
        if config is None:
            raise ValueError(f"No such transpiler: {transpiler_name}")
        return config.path

    def transpiler_config_options(self, transpiler_name: str, source_dialect: str) -> Sequence[LSPConfigOptionV1]:
        """
        Query the additional configuration options that are available for a given transpiler and source dialect that it
        supports.

        Args:
          transpiler_name: The transpiler name of the transpiler that may provide options.
          source_dialect: The source dialect supported by the given for which options are sought.
        Returns:
          A sequence of configuration options, possibly empty, that are supported by the given transpiler for the
          specified source dialect.
        """
        config = self._find_transpile_config(transpiler_name)
        if config is None:
            return []  # gracefully returns an empty list, since this can only happen during testing
        return config.options_for_dialect(source_dialect)

    def _all_transpiler_configs(self) -> Iterable[tuple[Path, LSPConfig]]:
        for path in self._transpiler_locations():
            config = self._transpiler_config(path)
            if config:
                yield path, config

    @classmethod
    def _transpiler_config(cls, path: Path) -> LSPConfig | None:
        if not path.is_dir() or not (path / "lib").is_dir():
            return None
        config_path = path / "lib" / "config.yml"
        if not config_path.is_file():
            return None
        try:
            return LSPConfig.load(config_path)
        except ValueError as e:
            logger.error(f"Could not load config: {path!s}", exc_info=e)
            return None
